import os
import tempfile
import pytest
from click.testing import CliRunner
from mpralib.cli import cli


@pytest.fixture(scope="module")
def runner():

    return CliRunner()


def test_combine_group_exists(runner):
    result = runner.invoke(cli, ["plot", "--help"])
    assert result.exit_code == 0
    assert "Plotting functions." in result.output


def test_combine_get_counts_option_required(runner):
    result = runner.invoke(cli, ["plot", "barcodes-per-oligo"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


def test_combine_get_reporter_elements_option_required(runner):
    result = runner.invoke(cli, ["plot", "correlation"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


def test_combine_get_reporter_variants_option_required(runner):
    result = runner.invoke(cli, ["plot", "dna-vs-rna"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


def test_combine_get_reporter_genomic_elements_option_required(runner):
    result = runner.invoke(cli, ["plot", "outlier"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


@pytest.fixture
def files():
    base = os.path.dirname(__file__)
    output_file = tempfile.NamedTemporaryFile(delete=False).name
    yield {
        "input": {
            "barcode_counts": os.path.join(base, "data", "reporter_experiment_barcode_IGVFDS2165KBMD.input.head20000.tsv.gz"),
        },
        "output": output_file,
    }
    os.remove(output_file)


def test_plot_barcodes_per_oligo(runner, files):
    result = runner.invoke(
        cli,
        [
            "plot",
            "barcodes-per-oligo",
            "--input",
            files["input"]["barcode_counts"],
            "--output",
            files["output"],
        ],
    )

    assert result.exit_code == 0
    assert os.path.exists(files["output"])


def test_plot_correlation(runner, files):
    result = runner.invoke(
        cli,
        [
            "plot",
            "correlation",
            "--input",
            files["input"]["barcode_counts"],
            "--oligos",
            "--bc-threshold",
            10,
            "--modality",
            "activity",
            "--output",
            files["output"],
        ],
    )

    assert result.exit_code == 0
    assert os.path.exists(files["output"])


def test_plot_outlier(runner, files):
    result = runner.invoke(
        cli,
        [
            "plot",
            "outlier",
            "--input",
            files["input"]["barcode_counts"],
            "--bc-threshold",
            2,
            "--output",
            files["output"],
        ],
    )

    assert result.exit_code == 0
    assert os.path.exists(files["output"])


def test_plot_dna_vs_rna(runner, files):
    result = runner.invoke(
        cli,
        [
            "plot",
            "outlier",
            "--input",
            files["input"]["barcode_counts"],
            "--bc-threshold",
            2,
            "--output",
            files["output"],
        ],
    )

    assert result.exit_code == 0
    assert os.path.exists(files["output"])



import os
import tempfile
import gzip
import pytest
from click.testing import CliRunner
from mpralib.cli import cli


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


def test_functional_group_help(runner):
    result = runner.invoke(cli, ["functional", "--help"])
    assert result.exit_code == 0
    print(result.output)
    assert "General functionality." in result.output


def test_functional_filter_outliers_help(runner):
    result = runner.invoke(cli, ["functional", "filter", "--help"])
    assert result.exit_code == 0
    print(result.output)
    assert "Usage: cli functional filter [OPTIONS]" in result.output


@pytest.fixture
def files():
    input_file = os.path.join(os.path.dirname(__file__), "data", "reporter_experiment_barcode.input.tsv.gz")
    output_file_activity = tempfile.NamedTemporaryFile(delete=False).name
    output_file_barcode = tempfile.NamedTemporaryFile(delete=False).name
    yield {"input": input_file, "output_activity": output_file_activity, "output_barcode": output_file_barcode}
    os.remove(output_file_activity)
    os.remove(output_file_barcode)


def test_filter_min_count_method(runner, files):
    # Should run without error and produce output files
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "min_count",
            "--method-values",
            '{"rna_min_count": 1}',
            "--bc-threshold",
            "1",
            "--output-activity",
            files["output_activity"],
            "--output-barcode",
            files["output_barcode"],
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists(files["output_activity"])
    assert os.path.exists(files["output_barcode"])


def test_filter_max_count_method(runner, files):
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "max_count",
            "--method-values",
            '{"dna_max_count": 10000}',
            "--output-activity",
            files["output_activity"],
            "--output-barcode",
            files["output_barcode"],
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists(files["output_activity"])
    assert os.path.exists(files["output_barcode"])


def test_filter_random_method(runner, files):
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "random",
            "--method-values",
            '{"proportion": 0.5}',
            "--output-activity",
            files["output_activity"],
            "--output-barcode",
            files["output_barcode"],
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists(files["output_barcode"])
    assert os.path.exists(files["output_activity"])


def test_filter_random_method_with_dict(runner, files):
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "random",
            "--method-values",
            "{'proportion': 0.5}",
            "--output-activity",
            files["output_activity"],
            "--output-barcode",
            files["output_barcode"],
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists(files["output_barcode"])
    assert os.path.exists(files["output_activity"])


def test_filter_random_method_with_invalid_json(runner, files):
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "random",
            "--method-values",
            "not_a_dict",
        ],
    )
    assert result.exit_code != 0
    assert "Could not parse --method-values as dict or JSON." in result.output


def test_filter_without_method_values(runner, files):
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "min_count",
            "--output-activity",
            files["output_activity"],
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists(files["output_activity"])


def test_functional_filter_outliers_global(runner, files):

    # Run the command
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "global",
            "--method-values",
            '{"times_zscore": 1000}',
            "--output-activity",
            files["output_activity"],
            "--output-barcode",
            files["output_barcode"],
        ],
    )

    # Check the result
    assert result.exit_code == 0
    assert os.path.exists(files["output_activity"])
    assert os.path.exists(files["output_barcode"])

    with open(files["output_barcode"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_experiment_barcode.input.tsv.gz")
    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content

    with open(files["output_activity"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_activity.bc1.output.tsv.gz")
    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content


def test_functional_filter_outliers_large_expression(runner, files):

    # Run the command
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "large_expression",
            "--output-activity",
            files["output_activity"],
            "--output-barcode",
            files["output_barcode"],
        ],
    )

    # Check the result
    assert result.exit_code == 0
    assert os.path.exists(files["output_activity"])
    assert os.path.exists(files["output_barcode"])

    with open(files["output_barcode"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_experiment_barcode.input.tsv.gz")
    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content

    with open(files["output_activity"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_activity.bc1.output.tsv.gz")
    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content


def test_functional_filter_outliers_large_expression_2(runner, files):

    # Run the command
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "large_expression",
            "--method-values",
            '{"times_activity": 2.0}',
            "--output-activity",
            files["output_activity"],
            "--output-barcode",
            files["output_barcode"],
        ],
    )

    # Check the result
    assert result.exit_code == 0
    assert os.path.exists(files["output_activity"])
    assert os.path.exists(files["output_barcode"])

    with open(files["output_barcode"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(
        os.path.dirname(__file__), "data", "functional", "reporter_experiment_barcode.filter.large_expression.output.tsv.gz"
    )
    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content

    with open(files["output_activity"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(
        os.path.dirname(__file__), "data", "functional", "reporter_experiment.filter.large_expression.output.tsv.gz"
    )
    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content


def test_functional_filter_dna_max_count_10(runner, files):

    # Run the command
    result = runner.invoke(
        cli,
        [
            "functional",
            "filter",
            "--input",
            files["input"],
            "--method",
            "max_count",
            "--method-values",
            '{"dna_max_count": 10}',
            "--output-activity",
            files["output_activity"],
            "--output-barcode",
            files["output_barcode"],
        ],
    )

    # Check the result
    assert result.exit_code == 0
    assert os.path.exists(files["output_activity"])
    assert os.path.exists(files["output_barcode"])

    with open(files["output_barcode"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(
        os.path.dirname(__file__), "data", "functional", "reporter_experiment_barcode.filter.max_count.output.tsv.gz"
    )
    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content

    with open(files["output_activity"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(
        os.path.dirname(__file__), "data", "functional", "reporter_experiment.filter.max_count.output.tsv.gz"
    )
    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content



import click
from typing import Optional
import logging
import pandas as pd
import numpy as np
import math
import pysam
from sklearn.preprocessing import MinMaxScaler
from mpralib.mpradata import MPRABarcodeData, BarcodeFilter, Modality
from mpralib.utils.io import (
    chromosome_map,
    export_activity_file,
    export_barcode_file,
    export_counts_file,
    read_sequence_design_file,
)
import mpralib.utils.plot as plt
from mpralib.utils.file_validation import validate_tsv_with_schema, ValidationSchema
import ast
import json
from builtins import filter as buildins_filter

pd.options.mode.copy_on_write = True


@click.group(help="Command line interface of MPRAlib, a library for MPRA data analysis.")
def cli() -> None:
    pass


@cli.group(help="Validate standardized MPRA reporter formats.")
def validate_file() -> None:
    pass


@validate_file.command(help="Validate MPRA reporter sequence design file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Sequence Design file to validate.",
)
def reporter_sequence_design(input_file: str) -> None:
    """
    Validate MPRA reporter sequence design file.

    Args:
        input_file (str): Path to the MPRA Reporter Sequence Design file to validate.
    """
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_SEQUENCE_DESIGN):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate MPRA Reporter Barcode to Element Mapping file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Barcode to Element Mapping file to validate.",
)
def reporter_barcode_to_element_mapping(input_file: str) -> None:
    """
    Validate MPRA Reporter Barcode to Element Mapping file.

    Args:
        input_file (str): Path to the MPRA Reporter Barcode to Element Mapping file to validate.
    """
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_BARCODE_TO_ELEMENT_MAPPING):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Experiment Barcode file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Experiment Barcode file to validate.",
)
def reporter_experiment_barcode(input_file: str) -> None:
    """
    Validate MPRA Reporter Experiment Barcode file.

    Args:
        input_file (str): Path to the MPRA Reporter Experiment Barcode file to validate.
    """
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_EXPERIMENT_BARCODE):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Experiment file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Experiment file to validate.",
)
def reporter_experiment(input_file: str) -> None:
    """
    Validate MPRA Reporter Experiment file.

    Args:
        input_file (str): Path to the MPRA Reporter Experiment file to validate.
    """
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_EXPERIMENT):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Element file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Element file to validate.",
)
def reporter_element(input_file: str) -> None:
    """
    Validate MPRA Reporter Element file.

    Args:
        input_file (str): Path to the MPRA Reporter Element file to validate.
    """
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_ELEMENT):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Variant file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Variant file to validate.",
)
def reporter_variant(input_file: str) -> None:
    """
    Validate MPRA Reporter Variant file.

    Args:
        input_file (str): Path to the MPRA Reporter Variant file to validate.
    """
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_VARIANT):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Genomic Element file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Genomic Element file to validate.",
)
def reporter_genomic_element(input_file: str) -> None:
    """
    Validate MPRA Reporter Genomic Element file.

    Args:
        input_file (str): Path to the MPRA Reporter Genomic Element file to validate.
    """
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_GENOMIC_ELEMENT):
        raise click.ClickException("Validation failed. Please check the input file.")


@validate_file.command(help="Validate Reporter Genomic Variant file.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="MPRA Reporter Genomic Variant file to validate.",
)
def reporter_genomic_variant(input_file: str) -> None:
    """
    Validate MPRA Reporter Genomic Variant file.

    Args:
        input_file (str): Path to the MPRA Reporter Genomic Variant file to validate.
    """
    if not validate_tsv_with_schema(input_file, ValidationSchema.REPORTER_GENOMIC_VARIANT):
        raise click.ClickException("Validation failed. Please check the input file.")


@cli.group(help="General functionality.")
def functional() -> None:
    pass


@functional.command(help="Generating element activity or barcode count files.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA result in a barcode format.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output (element level only).",
)
@click.option(
    "--element-level/--barcode-level",
    "element_level",
    default=True,
    help="Export activity at the element (default) or barcode level.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of results.",
)
def activities(input_file: str, bc_threshold: int, element_level: bool, output_file: str) -> None:
    """
    Generating element activity or barcode count files.

    Args:
        input_file (str): Input file path of MPRA result in a barcode format.
        bc_threshold (int): Using a barcode threshold for output (element level only).
        element_level (bool): Export activity at the element (default) or barcode level.
        output_file (str): Output file of results.
    """
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    if element_level:
        export_activity_file(mpradata.oligo_data, output_file)
    else:
        export_barcode_file(mpradata, output_file)


@functional.command(help="Compute pairwise correlations under a certain barcode threshold.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA result in a barcode format.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output (element level only).",
)
@click.option(
    "--correlation-method",
    "correlation_method",
    required=False,
    default="all",
    type=click.Choice(["pearson", "spearman", "all"]),
    help="Computing pearson, spearman or both.",
)
@click.option(
    "--correlation-on",
    "correlation_on",
    required=False,
    default="all",
    type=click.Choice(["dna_normalized", "rna_normalized", "activity", "all"]),
    help="Using a barcode threshold for output (element level only).",
)
def compute_correlation(input_file: str, bc_threshold: int, correlation_on: str, correlation_method: str) -> None:
    """
    Compute pairwise correlations under a certain barcode threshold.

    Args:
        input_file (str): Input file path of MPRA result in a barcode format.
        bc_threshold (int): Using a barcode threshold for output (element level only).
        correlation_on (str): Using a barcode threshold for output (element level only).
        correlation_method (str): Computing pearson, spearman or both.
    """
    mpradata = MPRABarcodeData.from_file(input_file).oligo_data

    mpradata.barcode_threshold = bc_threshold

    if correlation_on == "all":
        correlation_on_mode = [Modality.DNA_NORMALIZED, Modality.RNA_NORMALIZED, Modality.ACTIVITY]
    else:
        correlation_on_mode = [Modality.from_string(correlation_on)]

    if correlation_method == "all":
        correlation_method_list = ["pearson", "spearman"]
    else:
        correlation_method_list = [correlation_method]

    for method in correlation_method_list:
        for mode in correlation_on_mode:
            click.echo(f"{method} correlation on {mode}: {mpradata.correlation(method, mode).flatten()[[1, 2, 5]]}")


@functional.command(help="Filter barcodes based on min/max counts, random or detect barcode outliers.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--method",
    "method",
    required=True,
    type=click.Choice([f.name.lower() for f in BarcodeFilter]),
    help="Barcode filtering method. Choices: " + ", ".join([f.name.lower() for f in BarcodeFilter]),
)
@click.option(
    "--method-values",
    "method_values",
    required=False,
    type=str,
    help="JSON string or Python dict of parameters for the outlier detection method, e.g. '{\"times_zscore\": 3.0}'.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--output-activity",
    "output_activity_file",
    required=False,
    type=click.Path(writable=True),
    help="Output the activity file of results.",
)
@click.option(
    "--output-barcode",
    "output_barcode_file",
    required=False,
    type=click.Path(writable=True),
    help="Output the barcode file of results.",
)
def filter(
    input_file: str,
    method: str,
    method_values: str,
    bc_threshold: int,
    output_activity_file: Optional[str],
    output_barcode_file: Optional[str],
) -> None:
    """
    Filters barcodes from MPRA barcode data using different methods.

    Reads an input file to create an MPRAdata object, applies a barcode filter to remove outliers
    using the specified method and parameters, and optionally exports the filtered activity data
    to an output file. Prints Pearson correlation of log2FoldChange across replicates before and after outlier removal.

    Args:
        input_file (str): Path to the input file containing barcode data.
        method (BarcodeFilter): The method to use for outlier filtering.
        method_values (dict): Parameters for the outlier detection method.
        bc_threshold (int): Minimum barcode count threshold for generating the Pearson correlation.
        output_activity_file (str): Path to the output file to export filtered activity data. If None, no file is written.
        output_barcode_file (str): Path to the output file to export filtered barcode data. If None, no file is written.
    """

    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold
    oligo_data = mpradata.oligo_data

    click.echo(
        f"Pearson correlation log2FoldChange BEFORE filter: "
        f"{oligo_data.correlation('pearson', Modality.ACTIVITY).flatten()[[1, 2, 5]]}"
    )
    mpradata.var_filter = None
    # Parse method_values as dict if provided
    params = {}
    if method_values:
        try:
            params = json.loads(method_values)
        except Exception:
            try:
                params = ast.literal_eval(method_values)
            except Exception:
                raise click.ClickException("Could not parse --method-values as dict or JSON.")

    mpradata.apply_barcode_filter(BarcodeFilter.from_string(method), params)

    oligo_data = mpradata.oligo_data

    click.echo(
        f"Pearson correlation log2FoldChange AFTER filter: "
        f"{oligo_data.correlation('pearson', Modality.ACTIVITY).flatten()[[1, 2, 5]]}"
    )
    if output_activity_file:
        export_activity_file(oligo_data, output_activity_file)
    if output_barcode_file:
        export_barcode_file(mpradata, output_barcode_file)


@cli.group(help="Combine counts with other outputs.")
def combine():
    pass


@combine.command(help="Using a metadata file to generate a oligo to variant mapping.")
@click.option(
    "--input",
    "input_file",
    required=False,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results (barcode output file). "
    "If set only map of present oligos in the barcode count file will be generated.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of results.",
)
def get_variant_map(input_file: str, sequence_design_file: str, output_file: str) -> None:
    """
    Using a metadata file to generate a oligo to variant mapping.

    Args:
        input_file (str): Input file path of MPRA results (barcode output file).
        sequence_design_file (str): Sequence design file.
        output_file (str): Output file of results.
    """

    df_sequence_design = read_sequence_design_file(sequence_design_file)

    if input_file:
        print("Read barcode count file...")
        mpradata = MPRABarcodeData.from_file(input_file)

        print("Generating oligo data...")
        mpradata = mpradata.oligo_data

        print("Adding metadata file...")
        mpradata.add_sequence_design(df_sequence_design, sequence_design_file)

        print(mpradata.data.var["SPDI"])
        print(type(mpradata.data.var["SPDI"]))
        print(mpradata.data.var["SPDI"].values)

        print("Generating variant map...")
        variant_map = mpradata.variant_map
    else:
        variant_map = pd.DataFrame(
            {
                "ID": np.concatenate(df_sequence_design["SPDI"].values),
                "allele": np.concatenate(df_sequence_design["allele"].values),
                "oligo": df_sequence_design.index.values.repeat(df_sequence_design["SPDI"].apply(lambda x: len(x))),
            }
        )

        variant_map["REF"] = np.where(
            variant_map["allele"] == "ref", variant_map["oligo"], np.full(len(variant_map["allele"]), None, dtype=object)
        )
        variant_map["ALT"] = np.where(
            variant_map["allele"] == "alt", variant_map["oligo"], np.full(len(variant_map["allele"]), None, dtype=object)
        )
        variant_map = variant_map.groupby("ID").agg(
            {"REF": lambda x: list(buildins_filter(None, x)), "ALT": lambda x: list(buildins_filter(None, x))}
        )

    print("Prepair output file...")
    for key in ["REF", "ALT"]:
        # TODO: what happens a variant has multiple alt alleles?
        # Right now joined by comma. But maybe I hould use one row per ref/alt pai?
        variant_map.loc[:, key] = variant_map[key].apply(lambda x: ",".join(x))

    variant_map = variant_map[variant_map.apply(lambda x: all(x != ""), axis=1)]

    variant_map.to_csv(output_file, sep="\t", index=True)


@combine.command(help="Return DNA and RNA counts for all oligos or only thosed tagges as elements/ref snvs.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--oligos/--barcodes",
    "use_oligos",
    required=False,
    type=click.BOOL,
    default=True,
    help="Return counts per oligo or per barcode.",
)
@click.option(
    "--normalized-counts/--counts",
    "normalized_counts",
    required=False,
    type=click.BOOL,
    default=False,
    help="Getting counts or normalized counts.",
)
@click.option(
    "--elements-only/--all-oligos",
    "elements_only",
    required=False,
    type=click.BOOL,
    default=False,
    help="Only return count data for elements and ref sequence of variants.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of all non zero counts.",
)
def get_counts(
    input_file: str,
    sequence_design_file: str,
    bc_threshold: int,
    normalized_counts: bool,
    use_oligos: bool,
    elements_only: bool,
    output_file: str,
) -> None:
    """
    Return DNA and RNA counts for all oligos or only thosed tagges as elements/ref snvs

    Args:
        input_file (str): Input file path of MPRA results.
        sequence_design_file (str): Sequence design file.
        bc_threshold (int): Using a barcode threshold for output.
        normalized_counts (bool): Getting counts or normalized counts.
        use_oligos (bool): Return counts per oligo or per barcode.
        elements_only (bool): Only return count data for elements and ref sequence of variants.
        output_file (str): Output file of all non zero counts.
    """
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    if use_oligos:
        mpradata = mpradata.oligo_data

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    element_mask = None
    if elements_only:

        element_mask = np.array(mpradata.data.var["allele"].apply(lambda x: "ref" in x).values, dtype=bool) | (
            mpradata.data.var["category"] == "element"
        )
        element_mask = ~np.repeat(np.array(element_mask)[np.newaxis, :], mpradata.n_obs, axis=0)

    export_counts_file(mpradata, output_file, normalized=normalized_counts, filter=element_mask)


@combine.command(help="Write out DNA and RNA counts for REF and ALT oligos.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--normalized-counts/--counts",
    "normalized_counts",
    required=False,
    type=click.BOOL,
    default=False,
    help="Getting counts or normalized counts.",
)
@click.option(
    "--oligos/--barcodes",
    "use_oligos",
    required=False,
    type=click.BOOL,
    default=True,
    help="Return counst per oligo or per barcode.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of all non zero counts, divided by ref and alt.",
)
def get_variant_counts(
    input_file: str, sequence_design_file: str, bc_threshold: int, normalized_counts: bool, use_oligos: bool, output_file: str
) -> None:
    """Processes MPRA (Massively Parallel Reporter Assay) data to generate variant-level count tables.

    Args:
        input_file (str): Path to the input file containing barcode count data.
        sequence_design_file (str): Path to the sequence design file mapping barcodes to variants/oligos.
        bc_threshold (int): Minimum barcode count threshold for filtering.
        normalized_counts (bool): If True, use normalized counts; otherwise, use raw counts.
        use_oligos (bool): If True, aggregate counts at the oligo level and map to variants; otherwise, export counts directly.
        output_file (str): Path to the output file where the variant count table will be saved.

    Returns:
        None. The function writes the resulting count table to the specified output file.

    Notes:
        - When `use_oligos` is True, the function aggregates DNA and RNA counts for reference and alternate oligos
          per variant, across all replicates, and outputs a table with these counts.
        - When `use_oligos` is False, the function exports the counts using `export_counts_file`.
        - Variants with all-zero counts or counts only on reference or alternate are removed from the output.
    """
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    if use_oligos:
        mpradata = mpradata.oligo_data

        variant_map = mpradata.variant_map

        df = {"variant_id": []}
        for replicate in mpradata.obs_names:
            for rna_or_dna in ["dna", "rna"]:
                df[f"{rna_or_dna}_count_{replicate}_REF"] = []
                df[f"{rna_or_dna}_count_{replicate}_ALT"] = []

        if normalized_counts:
            dna_counts = mpradata.normalized_dna_counts.copy()
            rna_counts = mpradata.normalized_rna_counts.copy()
        else:
            dna_counts = mpradata.dna_counts.copy()
            rna_counts = mpradata.rna_counts.copy()

        not_observed_mask = ~mpradata.observed
        lower_bc_mask = mpradata.barcode_counts < mpradata.barcode_threshold
        not_variant_mask = ~np.repeat(
            np.array(mpradata.data.var["category"] == "variant")[np.newaxis, :], mpradata.n_obs, axis=0
        )

        dna_counts = np.ma.masked_array(dna_counts, mask=np.any([not_observed_mask, lower_bc_mask, not_variant_mask], axis=0))
        rna_counts = np.ma.masked_array(rna_counts, mask=np.any([not_observed_mask, lower_bc_mask, not_variant_mask], axis=0))

        for spdi, row in variant_map.iterrows():

            mask_ref = mpradata.oligos.isin(row["REF"])
            mask_alt = mpradata.oligos.isin(row["ALT"])

            df["variant_id"].append(spdi)

            dna_counts_ref = dna_counts[:, mask_ref].sum(axis=1)
            rna_counts_ref = rna_counts[:, mask_ref].sum(axis=1)

            dna_counts_alt = dna_counts[:, mask_alt].sum(axis=1)
            rna_counts_alt = rna_counts[:, mask_alt].sum(axis=1)

            for idx, replicate in enumerate(mpradata.obs_names):
                df[f"dna_count_{replicate}_REF"].append(dna_counts_ref[idx])
                df[f"rna_count_{replicate}_REF"].append(rna_counts_ref[idx])
                df[f"dna_count_{replicate}_ALT"].append(dna_counts_alt[idx])
                df[f"rna_count_{replicate}_ALT"].append(rna_counts_alt[idx])

        df = pd.DataFrame(df).set_index("variant_id")
        # remove IDs which are all zero or just on ref or alt
        df = df.map(lambda x: np.nan if isinstance(x, np.ma.core.MaskedConstant) else x)
        ref_columns = [f"dna_count_{replicate}_REF" for replicate in mpradata.obs_names] + [
            f"rna_count_{replicate}_REF" for replicate in mpradata.obs_names
        ]
        alt_columns = [f"dna_count_{replicate}_ALT" for replicate in mpradata.obs_names] + [
            f"rna_count_{replicate}_ALT" for replicate in mpradata.obs_names
        ]
        nan_columns = df.loc[:, ref_columns].isna().all(axis=1) | df.loc[:, alt_columns].isna().all(axis=1)
        df = df[~nan_columns]

        if normalized_counts:
            df.to_csv(output_file, sep="\t", index=True, na_rep="", float_format="%.6f")
        else:
            df.to_csv(output_file, sep="\t", index=True, na_rep="", float_format="%.0f")
    else:
        export_counts_file(
            mpradata,
            output_file,
            normalized=normalized_counts,
            filter=~np.repeat(np.array(mpradata.data.var["category"] == "variant")[np.newaxis, :], mpradata.n_obs, axis=0),
        )


@combine.command(help="Generate reporter element file from counts, sequence design and quantification statistics.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--statistics",
    "statistics_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file of mpralm or BCalm results.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--output-reporter-elements",
    "output_reporter_elements_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of MPRA data object.",
)
def get_reporter_elements(
    input_file: str, sequence_design_file: str, statistics_file: str, bc_threshold: int, output_reporter_elements_file: str
) -> None:
    """
    Generate reporter element file from counts, sequence design and quantification statistics.

    Args:
        input_file (str): Input file path of MPRA results.
        sequence_design_file (str): Sequence design file.
        statistics_file (str): Input file of mpralm or BCalm results.
        bc_threshold (int): Using a barcode threshold for output.
        output_reporter_elements_file (str): Output file of MPRA data object.
    """
    mpradata = MPRABarcodeData.from_file(input_file).oligo_data

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    df = pd.read_csv(statistics_file, sep="\t", header=0)

    indexes_in_order = [mpradata.oligos[mpradata.oligos == ID].index.tolist() for ID in df["ID"]]
    indexes_in_order = [index for sublist in indexes_in_order for index in sublist]
    df.index = pd.Index(indexes_in_order)

    df = df.join(mpradata.oligos, how="right")

    mpradata.data.varm["mpralm_element"] = df

    out_df = df[["oligo", "logFC", "P.Value", "adj.P.Val"]]
    out_df.loc[:, "inputCount"] = mpradata.normalized_dna_counts.mean(axis=0)
    out_df.loc[:, "outputCount"] = mpradata.normalized_rna_counts.mean(axis=0)
    out_df.dropna(inplace=True)
    out_df.loc[out_df["P.Value"] == 0, "P.Value"] = np.finfo(float).eps
    out_df.loc[out_df["adj.P.Val"] == 0, "adj.P.Val"] = np.finfo(float).eps
    out_df.loc[:, "minusLog10PValue"] = np.abs(np.log10(out_df.loc[:, "P.Value"]))
    out_df.loc[:, "minusLog10QValue"] = np.abs(np.log10(out_df.loc[:, "adj.P.Val"]))
    out_df = out_df.rename(columns={"oligo": "oligo_name", "logFC": "log2FoldChange"})
    out_df[
        [
            "oligo_name",
            "log2FoldChange",
            "inputCount",
            "outputCount",
            "minusLog10PValue",
            "minusLog10QValue",
        ]
    ].to_csv(output_reporter_elements_file, sep="\t", index=False, float_format="%.4f")


@combine.command(help="Generate reporter variant file from counts, sequence design and quantification statistics.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--statistics",
    "statistics_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file of the result from mpralm or BCalm.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--output-reporter-variants",
    "output_reporter_variants_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of MPRA data object.",
)
def get_reporter_variants(
    input_file: str, sequence_design_file: str, statistics_file: str, bc_threshold: int, output_reporter_variants_file: str
) -> None:
    """
    Generate reporter variant file from counts, sequence design and quantification statistics.

    Args:
        input_file (str): Input file path of MPRA results.
        sequence_design_file (str): Sequence design file.
        statistics_file (str): Input file of the result from mpralm or BCalm
        bc_threshold (int): Using a barcode threshold for output.
        output_reporter_variants_file (str): Output file of MPRA data object.
    """
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    mpradata = mpradata.oligo_data

    variant_map = mpradata.variant_map

    df = pd.read_csv(statistics_file, sep="\t", header=0, index_col=0, na_values="NA").dropna()

    dna_counts = mpradata.normalized_dna_counts.copy()
    rna_counts = mpradata.normalized_rna_counts.copy()

    for spdi, row in variant_map.iterrows():

        mask_ref = mpradata.oligos.isin(row["REF"])
        mask_alt = mpradata.oligos.isin(row["ALT"])

        dna_counts_ref = dna_counts[:, mask_ref].sum(axis=1)
        rna_counts_ref = rna_counts[:, mask_ref].sum(axis=1)

        dna_counts_alt = dna_counts[:, mask_alt].sum(axis=1)
        rna_counts_alt = rna_counts[:, mask_alt].sum(axis=1)

        if spdi in df.index:
            df.loc[str(spdi), "inputCountRef"] = dna_counts_ref.mean()
            df.loc[str(spdi), "inputCountAlt"] = dna_counts_alt.mean()
            df.loc[str(spdi), "outputCountRef"] = rna_counts_ref.mean()
            df.loc[str(spdi), "outputCountAlt"] = rna_counts_alt.mean()
            df.loc[str(spdi), "variantPos"] = int(
                mpradata.data.var["variant_pos"][mpradata.oligos.isin(row["REF"])].values[0][0]
            )

    df.loc[df["P.Value"] == 0, "P.Value"] = np.finfo(float).eps
    df.loc[df["adj.P.Val"] == 0, "adj.P.Val"] = np.finfo(float).eps
    df["minusLog10PValue"] = np.abs(np.log10(df["P.Value"]))
    df["minusLog10QValue"] = np.abs(np.log10(df["adj.P.Val"]))
    df.rename(
        columns={
            "CI.L": "CI_lower_95",
            "CI.R": "CI_upper_95",
            "logFC": "log2FoldChange",
        },
        inplace=True,
    )
    df["postProbEffect"] = df["B"].apply(lambda x: math.exp(x) / (1 + math.exp(x)))
    df["variant_id"] = df.index
    df["refAllele"] = df["variant_id"].apply(lambda x: x.split(":")[2])
    df["altAllele"] = df["variant_id"].apply(lambda x: x.split(":")[3])
    df["variantPos"] = df["variantPos"].astype(int)

    df[
        [
            "variant_id",
            "log2FoldChange",
            "inputCountRef",
            "outputCountRef",
            "inputCountAlt",
            "outputCountAlt",
            "minusLog10PValue",
            "minusLog10QValue",
            "postProbEffect",
            "CI_lower_95",
            "CI_upper_95",
            "variantPos",
            "refAllele",
            "altAllele",
        ]
    ].to_csv(output_reporter_variants_file, sep="\t", index=False, float_format="%.4f")


@combine.command(help="Generate reporter genomic elements file from counts, sequence design and quantification statistics.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--statistics",
    "statistics_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA lm file.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--reference",
    "reference",
    required=True,
    type=str,
    help="Using only this reference as denoted in ref in the sequence design file.",
)
@click.option(
    "--output-reporter-genomic-elements",
    "output_reporter_genomic_elements_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of MPRA data object.",
)
def get_reporter_genomic_elements(
    input_file: str,
    sequence_design_file: str,
    statistics_file: str,
    bc_threshold: int,
    reference: str,
    output_reporter_genomic_elements_file: str,
) -> None:
    """
    Generate reporter genomic elements file from counts, sequence design and quantification statistics.

    Args:
        input_file (str): Input file path of MPRA results.
        sequence_design_file (str): Sequence design file.
        statistics_file (str): Input file path of MPRA lm file.
        bc_threshold (int): Using a barcode threshold for output.
        reference (str): Using only this reference as denoted in ref in the sequence design file
        output_reporter_genomic_elements_file (str): Output file of MPRA data object.
    """
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    mpradata = mpradata.oligo_data

    mask = np.array(mpradata.data.var["allele"].apply(lambda x: "ref" in x).values, dtype=bool) | (
        mpradata.data.var["category"] == "element"
    )
    mask = mask & (mpradata.data.var["ref"] == reference)

    df = pd.read_csv(statistics_file, sep="\t", header=0)

    indexes_in_order = [mpradata.oligos[mpradata.oligos == ID].index.tolist() for ID in df["ID"]]
    indexes_in_order = [index for sublist in indexes_in_order for index in sublist]
    df.index = pd.Index(indexes_in_order)

    df = df.join(mpradata.oligos, how="right")

    mpradata.data.varm["mpralm_element"] = df

    out_df = df[["oligo", "logFC", "P.Value", "adj.P.Val"]]

    out_df.loc[:, ["inputCount"]] = mpradata.normalized_dna_counts.mean(axis=0)
    out_df.loc[:, ["outputCount"]] = mpradata.normalized_rna_counts.mean(axis=0)

    out_df["chr"] = mpradata.data.var["chr"]
    out_df["start"] = mpradata.data.var["start"]
    out_df["end"] = mpradata.data.var["end"]
    out_df["strand"] = mpradata.data.var["strand"]

    # apply ref and element mask
    out_df = out_df[mask]

    out_df.dropna(inplace=True)
    scaler = MinMaxScaler(feature_range=(0, 1000))
    out_df["score"] = scaler.fit_transform(np.abs(out_df[["logFC"]])).astype(int)
    out_df.loc[out_df["P.Value"] == 0, "P.Value"] = np.finfo(float).eps
    out_df.loc[out_df["adj.P.Val"] == 0, "adj.P.Val"] = np.finfo(float).eps
    out_df["minusLog10PValue"] = np.abs(np.log10(out_df["P.Value"]))
    out_df["minusLog10QValue"] = np.abs(np.log10(out_df["adj.P.Val"]))
    out_df.rename(columns={"oligo": "name", "logFC": "log2FoldChange"}, inplace=True)
    out_df = out_df[
        [
            "chr",
            "start",
            "end",
            "name",
            "score",
            "strand",
            "log2FoldChange",
            "inputCount",
            "outputCount",
            "minusLog10PValue",
            "minusLog10QValue",
        ]
    ].sort_values(by=["chr", "start", "end"])

    with pysam.BGZFile(output_reporter_genomic_elements_file, "ab", None) as f:
        f.write(out_df.to_csv(sep="\t", index=False, header=False, float_format="%.4f").encode())


@combine.command(help="Generate reporter genomic variants file from counts, sequence design and quantification statistics.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--sequence-design",
    "sequence_design_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Sequence design file.",
)
@click.option(
    "--statistics",
    "statistics_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file of the result from mpralm or BCalm.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--reference",
    "reference",
    required=True,
    default="GRCh38",
    type=click.Choice(["GRCh38", "GRCh37"]),
    help="Using only this reference as denoted in ref in the sequence design file.",
)
@click.option(
    "--output-reporter-genomic-variants",
    "output_reporter_genomic_variants_file",
    required=True,
    type=click.Path(writable=True),
    help="Output file of MPRA data object.",
)
def get_reporter_genomic_variants(
    input_file: str,
    sequence_design_file: str,
    statistics_file: str,
    bc_threshold: int,
    reference: str,
    output_reporter_genomic_variants_file: str,
) -> None:
    """
    Generate reporter genomic variants file from counts, sequence design and quantification statistics.

    Args:
        input_file (str): Input file path of MPRA results.
        sequence_design_file (str): Sequence design file.
        statistics_file (str): Input file of the result from mpralm or BCalm
        bc_threshold (int): Using a barcode threshold for output.
        reference (str): Using only this reference as denoted in ref in the sequence design file
        output_reporter_genomic_variants_file (str): Output file of MPRA data object
    """
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.add_sequence_design(read_sequence_design_file(sequence_design_file), sequence_design_file)

    mpradata.barcode_threshold = bc_threshold

    mpradata = mpradata.oligo_data

    variant_map = mpradata.variant_map

    df = pd.read_csv(statistics_file, sep="\t", header=0, index_col=0, na_values="NA").dropna()

    dna_counts = mpradata.normalized_dna_counts.copy()
    rna_counts = mpradata.normalized_rna_counts.copy()

    for spdi, row in variant_map.iterrows():

        mask_ref = mpradata.oligos.isin(row["REF"])
        mask_alt = mpradata.oligos.isin(row["ALT"])

        dna_counts_ref = dna_counts[:, mask_ref].sum(axis=1)
        rna_counts_ref = rna_counts[:, mask_ref].sum(axis=1)

        dna_counts_alt = dna_counts[:, mask_alt].sum(axis=1)
        rna_counts_alt = rna_counts[:, mask_alt].sum(axis=1)

        if spdi in df.index:
            df.loc[str(spdi), "inputCountRef"] = dna_counts_ref.mean()
            df.loc[str(spdi), "inputCountAlt"] = dna_counts_alt.mean()
            df.loc[str(spdi), "outputCountRef"] = rna_counts_ref.mean()
            df.loc[str(spdi), "outputCountAlt"] = rna_counts_alt.mean()
            df.loc[str(spdi), "variantPos"] = int(
                mpradata.data.var["variant_pos"][mpradata.oligos.isin(row["REF"])].values[0][0]
            )
            df.loc[str(spdi), "strand"] = mpradata.data.var["strand"][mpradata.oligos.isin(row["REF"])].values[0]

    df["variantPos"] = df["variantPos"].astype(int)
    df.loc[df["P.Value"] == 0, "P.Value"] = np.finfo(float).eps
    df.loc[df["adj.P.Val"] == 0, "adj.P.Val"] = np.finfo(float).eps
    df["minusLog10PValue"] = np.abs(np.log10(df["P.Value"]))
    df["minusLog10QValue"] = np.abs(np.log10(df["adj.P.Val"]))
    df.rename(
        columns={
            "CI.L": "CI_lower_95",
            "CI.R": "CI_upper_95",
            "logFC": "log2FoldChange",
        },
        inplace=True,
    )
    df["postProbEffect"] = df["B"].apply(lambda x: math.exp(x) / (1 + math.exp(x)))
    df["variant_id"] = df.index
    df["refAllele"] = df["variant_id"].apply(lambda x: x.split(":")[2])
    df["altAllele"] = df["variant_id"].apply(lambda x: x.split(":")[3])
    df["start"] = df["variant_id"].apply(lambda x: x.split(":")[1]).astype(int)
    df["end"] = df["start"] + df["refAllele"].apply(lambda x: len(x)).astype(int)

    map = chromosome_map()
    map = map[map["release"] == reference]
    df["chr"] = df["variant_id"].apply(lambda x: _get_chr(map, x, mpradata.LOGGER))

    df.dropna(inplace=True)

    scaler = MinMaxScaler(feature_range=(0, 1000))
    df["score"] = scaler.fit_transform(np.abs(df[["log2FoldChange"]])).astype(int)

    df = df[
        [
            "chr",
            "start",
            "end",
            "variant_id",
            "score",
            "strand",
            "log2FoldChange",
            "inputCountRef",
            "outputCountRef",
            "inputCountAlt",
            "outputCountAlt",
            "minusLog10PValue",
            "minusLog10QValue",
            "postProbEffect",
            "CI_lower_95",
            "CI_upper_95",
            "variantPos",
            "refAllele",
            "altAllele",
        ]
    ].sort_values(by=["chr", "start", "end"])

    with pysam.BGZFile(output_reporter_genomic_variants_file, "ab", None) as f:
        f.write(df.to_csv(sep="\t", index=False, header=False, float_format="%.4f").encode())


def _get_chr(map: pd.DataFrame, variant_id: str, logger: logging.Logger) -> Optional[str]:
    variant_contig = variant_id.split(":")[0]
    if variant_contig in map["refseq"].values:
        return map[map["refseq"] == variant_contig].loc[:, "ucsc"].values[0]
    else:
        logger.warning(f"Contig {variant_contig} of SPDI {variant_id} not found in chromosome map. Returning None.")
        return None


@cli.group(help="Plotting functions.")
def plot() -> None:
    pass


@plot.command(help="Scatter plot to see correlations between replicates (activity, DNA or RNA).")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--oligos/--barcodes",
    "use_oligos",
    required=False,
    type=click.BOOL,
    default=True,
    help="Return counst per oligo or per barcode.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--modality",
    "modality",
    required=False,
    default="activity",
    type=click.Choice(["dna_normalized", "rna_normalized", "activity"]),
    help="What modality should be plotted.",
)
@click.option(
    "--replicate",
    "replicates",
    required=False,
    multiple=True,
    type=str,
    help="Compare only these two replicates. Otherwise all.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output plot file.",
)
def correlation(
    input_file: str, use_oligos: bool, bc_threshold: int, modality: str, replicates: Optional[list], output_file: str
) -> None:
    """
    Scatter plot to see correlations between replicates (activity, DNA or RNA).

    Reads a MPRA reporter experiment barcode file and generates a scatter plot to visualize the correlation between replicates. Can be done on barcode or oligo level as well as with a barcode count threshold (to get more confident correlation).

    Args:
        input_file (str): Path to the input file containing barcode data.
        use_oligos (bool): Whether to plot individual barcodes or aggregate over oligos.
        bc_threshold (int): Minimum barcode count threshold per oligo to plot.
        modality (str): The modality to use for the plot (e.g., "activity", "dna_normalized", "rna_normalized").
        replicates (list): List of replicate names to include in the plot. If None, all replicates are included.
        output_file (str): Path to the output file where the plot will be saved.
    """  # noqa: E501
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    if use_oligos:
        mpradata = mpradata.oligo_data

    mod = Modality.from_string(modality)

    if replicates:
        fig = plt.correlation(mpradata, mod, replicates)
    else:
        fig = plt.correlation(mpradata, mod)

    fig.savefig(output_file)


@plot.command(help="Plotting the DNA vs RNA counts (log10, median on  multiple replicates).")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--oligos/--barcodes",
    "use_oligos",
    required=False,
    type=click.BOOL,
    default=True,
    help="Return counst per oligo or per barcode.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--replicate",
    "replicates",
    required=False,
    multiple=True,
    type=str,
    help="Compare only these two replicates. Otherwise all.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output plot file.",
)
def dna_vs_rna(input_file: str, use_oligos: bool, bc_threshold: int, replicates: Optional[list], output_file: str) -> None:
    """
    Plotting the DNA vs RNA counts (log10, median on multiple replicates).

    Reads a MPRA reporter experiment barcode file and generates a scatter plot to visualize the DNA to RNA counts. This is usefull to see if you can actually see any activity in your data. Can be done on barcode or oligo level.

    Args:
        input_file (str): Path to the input file containing barcode data.
        use_oligos (bool): Whether to plot individual barcodes or aggregate over oligos.
        bc_threshold (int): Minimum barcode count threshold per oligo to plot.
        replicates (list): List of replicate names to compute the median on for the plot. If None, all replicates are included.
        output_file (str): Path to the output file where the plot will be saved.
    """  # noqa: E501
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    if use_oligos:
        mpradata = mpradata.oligo_data

    if replicates:
        fig = plt.dna_vs_rna(mpradata, replicates)
    else:
        fig = plt.dna_vs_rna(mpradata)

    fig.savefig(output_file)


@plot.command(help="Histogramm of barcodes per oligo.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--replicate",
    "replicates",
    required=False,
    multiple=True,
    type=str,
    help="Show only these replicates. Otheriwse all.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output plot file.",
)
def barcodes_per_oligo(input_file: str, replicates: Optional[list], output_file: str) -> None:
    """
    Histogramm of barcodes per oligo

    Generates a histogram of barcodes per oligo per replicate.

    Args:
        input_file (str): Path to the input file containing barcode data.
        replicates (list): List of replicate names to plot the histogram. If None, all replicates are included.
        output_file (str): Path to the output file where the plot will be saved.
    """  # noqa: E501
    mpradata = MPRABarcodeData.from_file(input_file).oligo_data

    if replicates:
        fig = plt.barcodes_per_oligo(mpradata, replicates)
    else:
        fig = plt.barcodes_per_oligo(mpradata)

    fig.savefig(output_file)


@plot.command(help="Outlier plot.")
@click.option(
    "--input",
    "input_file",
    required=True,
    type=click.Path(exists=True, readable=True),
    help="Input file path of MPRA results.",
)
@click.option(
    "--bc-threshold",
    "bc_threshold",
    required=False,
    default=1,
    type=int,
    help="Using a barcode threshold for output.",
)
@click.option(
    "--output",
    "output_file",
    required=True,
    type=click.Path(writable=True),
    help="Output plot file.",
)
def outlier(input_file: str, bc_threshold: int, output_file: str) -> None:
    """
    Outlier plot.

    Reads a MPRA reporter experiment barcode file and generates an outlier plot to visualize potential outlier samples based on barcode counts.

    Args:
        input_file (str): Path to the input file containing barcode data.
        bc_threshold (int): Minimum barcode count threshold per oligo to consider.
        output_file (str): Path to the output file where the plot will be saved.
    """  # noqa: E501
    mpradata = MPRABarcodeData.from_file(input_file)

    mpradata.barcode_threshold = bc_threshold

    fig = plt.barcodes_outlier(mpradata)

    fig.savefig(output_file)


if __name__ == "__main__":
    cli()



import numpy as np
import pandas as pd
import anndata as ad
import copy
import pytest
from mpralib.mpradata import MPRABarcodeData, CountSampling, BarcodeFilter, MPRAData, Modality, MPRAOligoData
from mpralib.mpradata import MPRAlibException


OBS = pd.DataFrame(index=["rep1", "rep2", "rep3"])
VAR = pd.DataFrame(
    {"oligo": ["oligo1", "oligo1", "oligo2", "oligo3", "oligo3"]},
    index=["barcode1", "barcode2", "barcode3", "barcode4", "barcode5"],
)
COUNTS_DNA = np.array([[1, 2, 3, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])
COUNTS_RNA = np.array([[1, 2, 4, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])

FILTER = np.array(
    [
        [False, True, False],
        [False, False, False],
        [True, False, False],
        [False, False, True],
        [False, False, True],
    ]
)


@pytest.fixture
def mpra_data():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    return MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))


@pytest.fixture
def mpra_data_with_bc_filter(mpra_data):
    data = copy.deepcopy(mpra_data)
    data.var_filter = FILTER
    return data


def test_apply_count_sampling_rna(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA, proportion=0.5)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    assert np.all(rna_sampling <= np.asarray(mpra_data.data.layers["rna"]))
    assert np.all(rna_sampling >= 0)


def test_apply_count_sampling_dna(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.DNA, proportion=0.5)
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.all(dna_sampling <= mpra_data.data.layers["dna"])
    assert np.all(dna_sampling >= 0)


def test_apply_count_sampling_rna_and_dna(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, proportion=0.5)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.all(rna_sampling <= mpra_data.data.layers["rna"])
    assert np.all(rna_sampling >= 0)
    assert np.all(dna_sampling <= np.asarray(mpra_data.data.layers["dna"]))
    assert np.all(dna_sampling >= 0)


def test_apply_count_sampling_rna_total(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA, total=10)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    assert np.sum(rna_sampling) <= 30


def test_apply_count_sampling_total(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, total=10)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.sum(rna_sampling) <= 30
    assert np.sum(dna_sampling) <= 30


def test_apply_count_sampling_max_value_rna(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA, max_value=2)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    assert np.all(rna_sampling <= 2)


def test_apply_count_sampling_max_value(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, max_value=2)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.all(rna_sampling <= 2)
    assert np.all(dna_sampling <= 2)


def test_apply_count_sampling_aggregate_over_replicates(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, total=10, aggregate_over_replicates=True)
    rna_sampling = np.asarray(mpra_data.data.layers["rna_sampling"])
    dna_sampling = np.asarray(mpra_data.data.layers["dna_sampling"])
    assert np.sum(rna_sampling) <= 11
    assert np.sum(dna_sampling) <= 11


def test_barcode_counts(mpra_data):
    supporting_barcodes = mpra_data.barcode_counts
    expected_barcodes = np.array([[2, 2, 1, 2, 2], [2, 2, 1, 2, 2], [2, 2, 1, 2, 2]])
    np.testing.assert_array_equal(supporting_barcodes, expected_barcodes)


def test_barcode_counts_with_filter(mpra_data_with_bc_filter):
    supporting_barcodes = mpra_data_with_bc_filter.barcode_counts
    expected_barcodes = np.array([[2, 2, 1, 2, 2], [2, 2, 1, 2, 2], [2, 2, 1, 2, 2]])
    np.testing.assert_array_equal(supporting_barcodes, expected_barcodes)


def test_raw_dna_counts(mpra_data):
    expected_dna_counts = np.array([[1, 2, 3, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])
    np.testing.assert_array_equal(mpra_data.raw_dna_counts, expected_dna_counts)


def test_raw_dna_counts_with_modification(mpra_data):
    mpra_data.data.layers["dna"] = np.array([[10, 20, 30, 10, 20], [40, 50, 60, 40, 50], [70, 80, 90, 100, 1000]])
    expected_dna_counts = np.array([[10, 20, 30, 10, 20], [40, 50, 60, 40, 50], [70, 80, 90, 100, 1000]])
    np.testing.assert_array_equal(mpra_data.raw_dna_counts, expected_dna_counts)


def test_filtered_dna_counts(mpra_data, mpra_data_with_bc_filter):
    expected_dna_counts = np.array([[1, 2, 3, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])
    np.testing.assert_array_equal(mpra_data.dna_counts, expected_dna_counts)
    expected_filtered_dna_counts = np.array([[1, 2, 0, 1, 2], [0, 5, 6, 4, 5], [7, 8, 9, 0, 0]])
    np.testing.assert_array_equal(mpra_data_with_bc_filter.dna_counts, expected_filtered_dna_counts)


def test_dna_counts_with_sampling(mpra_data):
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.DNA, proportion=0.5)
    dna_sampling = mpra_data.data.layers["dna_sampling"]
    np.testing.assert_array_equal(mpra_data.dna_counts, dna_sampling)


def test_dna_counts_with_filter(mpra_data, mpra_data_with_bc_filter):
    mpra_data.apply_count_sampling(CountSampling.DNA, max_value=2)
    expected_filtered_dna_counts = np.array([[1, 2, 2, 1, 2], [2, 2, 2, 2, 2], [2, 2, 2, 2, 2]])
    np.testing.assert_array_equal(mpra_data.dna_counts, expected_filtered_dna_counts)
    mpra_data_with_bc_filter.apply_count_sampling(CountSampling.DNA, max_value=2)
    expected_filtered_dna_counts = np.array([[1, 2, 0, 1, 2], [0, 2, 2, 2, 2], [2, 2, 2, 0, 0]])
    np.testing.assert_array_equal(mpra_data_with_bc_filter.dna_counts, expected_filtered_dna_counts)


@pytest.fixture
def mpra_data_barcode():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    return MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))


def test_apply_barcode_filter_min_count(mpra_data_barcode):
    mpra_data_barcode.apply_barcode_filter(BarcodeFilter.MIN_COUNT, params={"rna_min_count": 4, "dna_min_count": 3})
    expected_filter = np.array(
        [
            [True, False, False],
            [True, False, False],
            [False, False, False],
            [True, False, False],
            [True, False, False],
        ]
    )
    np.testing.assert_array_equal(mpra_data_barcode.var_filter, expected_filter)


def test_barcode_filter_other_to_min_max_counts(mpra_data_barcode):

    result = mpra_data_barcode._barcode_filter_min_max_counts(
        BarcodeFilter.LARGE_EXPRESSION, counts=mpra_data_barcode.raw_dna_counts, count_threshold=1
    )
    expected_filter = np.array(
        [
            [False, False, False],
            [False, False, False],
            [False, False, False],
            [False, False, False],
            [False, False, False],
        ]
    )
    np.testing.assert_array_equal(result, expected_filter)


def test_apply_barcode_filter_max_count(mpra_data_barcode):
    mpra_data_barcode.apply_barcode_filter(BarcodeFilter.MAX_COUNT, params={"rna_max_count": 9, "dna_max_count": 100})
    expected_filter = np.array(
        [
            [False, False, False],
            [False, False, False],
            [False, False, False],
            [False, False, True],
            [False, False, True],
        ]
    )
    np.testing.assert_array_equal(mpra_data_barcode.var_filter, expected_filter)

    mpra_data_barcode.var_filter = None
    mpra_data_barcode.apply_barcode_filter(BarcodeFilter.MAX_COUNT, params={"dna_max_count": 99})
    expected_filter = np.array(
        [
            [False, False, False],
            [False, False, False],
            [False, False, False],
            [False, False, False],
            [False, False, True],
        ]
    )
    np.testing.assert_array_equal(mpra_data_barcode.var_filter, expected_filter)  # type: ignore


@pytest.fixture
def mpra_data_norm():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    data = MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))
    data.scaling = 10
    return data


@pytest.fixture
def mpra_data_norm_with_bc_filter(mpra_data_norm):
    data = copy.deepcopy(mpra_data_norm)
    data.var_filter = FILTER
    return data


def test_normalize_counts(mpra_data_norm):
    mpra_data_norm._normalize()
    dna_normalized = mpra_data_norm.normalized_dna_counts
    expected_dna_normalized = np.array(
        [
            [1.428, 2.142, 2.857, 1.428, 2.142],
            [1.724, 2.068, 2.413, 1.724, 2.068],
            [0.575, 0.647, 0.719, 0.791, 7.266],
        ]
    )
    np.testing.assert_almost_equal(dna_normalized, expected_dna_normalized, decimal=3)
    expected_rna_normalized = np.array(
        [[1.333, 2.0, 3.333, 1.333, 2.0], [1.724, 2.069, 2.414, 1.724, 2.069], [0.576, 0.647, 0.719, 0.791, 7.266]]
    )
    rna_normalized = mpra_data_norm.normalized_rna_counts
    np.testing.assert_almost_equal(rna_normalized, expected_rna_normalized, decimal=3)


def test_normalize_without_pseudocount(mpra_data_norm):
    mpra_data = copy.deepcopy(mpra_data_norm)
    mpra_data.pseudo_count = 0
    mpra_data._normalize()
    dna_normalized = np.asarray(mpra_data.data.layers["dna_normalized"])
    expected_dna_normalized = np.array(
        [[1.111, 2.222, 3.333, 1.111, 2.222], [1.667, 2.083, 2.5, 1.667, 2.083], [0.522, 0.597, 0.672, 0.746, 7.463]]
    )
    np.testing.assert_almost_equal(dna_normalized, expected_dna_normalized, decimal=3)
    expected_rna_normalized = np.array(
        [[1.0, 2.0, 4.0, 1.0, 2.0], [1.667, 2.083, 2.5, 1.667, 2.083], [0.522, 0.597, 0.672, 0.746, 7.463]]
    )
    rna_normalized = np.asarray(mpra_data.data.layers["rna_normalized"])
    np.testing.assert_almost_equal(rna_normalized, expected_rna_normalized, decimal=3)


def test_normalize_counts_with_bc_filter(mpra_data_norm_with_bc_filter):
    mpra_data_norm_with_bc_filter._normalize()

    dna_normalized = np.asarray(mpra_data_norm_with_bc_filter.data.layers["dna_normalized"])
    expected_normalized = np.array(
        [
            [1.428, 2.142, 2.857, 1.428, 2.142],
            [1.724, 2.068, 2.413, 1.724, 2.068],
            [0.575, 0.647, 0.719, 0.791, 7.266],
        ]
    )
    np.testing.assert_almost_equal(dna_normalized, expected_normalized, decimal=3)

    dna_normalized = mpra_data_norm_with_bc_filter.normalized_dna_counts
    expected_normalized = np.array(
        [
            [1.428, 2.142, 0.0, 1.428, 2.142],
            [0.0, 2.068, 2.413, 1.724, 2.068],
            [0.575, 0.647, 0.719, 0.0, 0.0],
        ]
    )
    np.testing.assert_almost_equal(dna_normalized, expected_normalized, decimal=3)

    rna_normalized = np.asarray(mpra_data_norm_with_bc_filter.data.layers["rna_normalized"])
    expected_normalized = np.array(
        [[1.333, 2.0, 3.333, 1.333, 2.0], [1.724, 2.069, 2.414, 1.724, 2.069], [0.576, 0.647, 0.719, 0.791, 7.266]]
    )
    np.testing.assert_almost_equal(rna_normalized, expected_normalized, decimal=3)
    rna_normalized = mpra_data_norm_with_bc_filter.normalized_rna_counts
    expected_normalized = np.array(
        [[1.333, 2.0, 0.0, 1.333, 2.0], [0.0, 2.069, 2.414, 1.724, 2.069], [0.576, 0.647, 0.719, 0.0, 0.0]]
    )
    np.testing.assert_almost_equal(rna_normalized, expected_normalized, decimal=3)


@pytest.fixture
def mpra_oligo_data():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    mpra_barcode_data = MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))
    data = mpra_barcode_data.oligo_data
    data.scaling = 10
    return data


@pytest.fixture
def mpra_oligo_data_with_bc_filter():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    mpra_barcode_data = MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))
    mpra_barcode_data.var_filter = FILTER
    data = mpra_barcode_data.oligo_data
    data.scaling = 10
    return data


def test_oligo_normalize_counts(mpra_oligo_data):
    dna_normalized = mpra_oligo_data.normalized_dna_counts
    expected_dna_normalized = np.array([[1.786, 2.857, 1.786], [1.897, 2.414, 1.897], [0.612, 0.719, 4.029]])
    np.testing.assert_almost_equal(dna_normalized, expected_dna_normalized, decimal=3)
    expected_rna_normalized = np.array([[1.667, 3.333, 1.667], [1.897, 2.414, 1.897], [0.612, 0.719, 4.029]])
    rna_normalized = mpra_oligo_data.normalized_rna_counts
    np.testing.assert_almost_equal(rna_normalized, expected_rna_normalized, decimal=3)


def test_oligo_normalize_without_pseudocount(mpra_oligo_data):
    mpra_data = copy.deepcopy(mpra_oligo_data)
    mpra_data.pseudo_count = 0
    dna_normalized = mpra_data.normalized_dna_counts
    expected_dna_normalized = np.array([[1.667, 3.333, 1.667], [1.875, 2.5, 1.875], [0.56, 0.672, 4.104]])
    np.testing.assert_almost_equal(dna_normalized, expected_dna_normalized, decimal=3)
    expected_rna_normalized = np.array([[1.5, 4.0, 1.5], [1.875, 2.5, 1.875], [0.56, 0.672, 4.104]])
    rna_normalized = mpra_data.normalized_rna_counts
    np.testing.assert_almost_equal(rna_normalized, expected_rna_normalized, decimal=3)


def test_oligo_normalize_counts_with_bc_filter(mpra_oligo_data_with_bc_filter):
    dna_normalized = mpra_oligo_data_with_bc_filter.normalized_dna_counts
    expected_normalized = np.array([[2.5, 0.0, 2.5], [2.5, 2.917, 2.292], [3.148, 3.704, 0.0]])
    np.testing.assert_almost_equal(dna_normalized, expected_normalized, decimal=3)
    rna_normalized = mpra_oligo_data_with_bc_filter.normalized_rna_counts
    np.testing.assert_almost_equal(rna_normalized, expected_normalized, decimal=3)


@pytest.fixture
def mpra_corr_data():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    return MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers)).oligo_data


def test_correlation(mpra_corr_data):
    mpra_corr_data._compute_correlation(mpra_corr_data.activity, "log2FoldChange")
    assert "pearson_correlation_log2FoldChange" in mpra_corr_data.data.obsp
    assert "spearman_correlation_log2FoldChange" in mpra_corr_data.data.obsp


def test_pearson_correlation(mpra_corr_data):
    x = mpra_corr_data.correlation(method="pearson", count_type=Modality.ACTIVITY)
    y = mpra_corr_data.correlation(method="pearson", count_type=Modality.RNA_NORMALIZED)
    z = mpra_corr_data.correlation(method="pearson", count_type=Modality.DNA_NORMALIZED)
    np.testing.assert_equal(x, np.array([[1.0, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]))
    np.testing.assert_almost_equal(
        y, np.array([[1.0, 1.0, -0.475752], [1.0, 1.0, -0.475752], [-0.475752, -0.475752, 1.0]]), decimal=3
    )
    np.testing.assert_almost_equal(z, np.array([[1.0, 1.0, -0.476], [1.0, 1.0, -0.476], [-0.476, -0.476, 1.0]]), decimal=3)


def test_spearman_correlation(mpra_corr_data):
    x = mpra_corr_data.correlation(method="spearman", count_type=Modality.ACTIVITY)
    y = mpra_corr_data.correlation(method="spearman", count_type=Modality.RNA_NORMALIZED)
    z = mpra_corr_data.correlation(method="spearman", count_type=Modality.DNA_NORMALIZED)
    np.testing.assert_equal(x, np.array([[1.0, np.nan, np.nan], [np.nan, np.nan, np.nan], [np.nan, np.nan, np.nan]]))
    np.testing.assert_almost_equal(y, np.array([[1.0, 1.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]]), decimal=3)
    np.testing.assert_equal(z, np.array([[1.0, 1.0, 0.0], [1.0, 1.0, 0.0], [0.0, 0.0, 1.0]]))


@pytest.fixture
def mpra_complexity_data():
    counts_dna = np.array([[0, 2, 0, 1, 2], [4, 5, 6, 4, 5], [7, 0, 9, 10, 0]])
    counts_rna = np.array([[1, 2, 0, 1, 2], [4, 5, 6, 4, 5], [7, 0, 9, 10, 0]])
    layers = {"rna": counts_rna, "dna": counts_dna}
    return MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))


def test_lincoln_complexity(mpra_complexity_data):
    complexity = mpra_complexity_data.complexity()
    np.testing.assert_equal(complexity, np.array([[4, 5, 6], [5, 5, 5], [6, 5, 3]]))


def test_chapman_complexity(mpra_complexity_data):
    complexity = mpra_complexity_data.complexity(method="chapman")
    np.testing.assert_equal(complexity, np.array([[4, 5, 5], [5, 5, 5], [5, 5, 3]]))


def test_fail_complexity(mpra_complexity_data):
    with pytest.raises(ValueError):
        mpra_complexity_data.complexity(method="unknown")


def test_read_and_write(tmp_path, mpra_data):
    out_path = tmp_path / "bc_data.h5ad"
    mpra_data.write(out_path)

    data = MPRABarcodeData.read(out_path)
    assert isinstance(data, MPRABarcodeData)
    assert data.data.shape == mpra_data.data.shape
    assert data.data.layers.keys() == mpra_data.data.layers.keys()
    assert np.all(data.rna_counts == mpra_data.rna_counts)
    assert np.all(data.activity == mpra_data.activity)
    assert data.pseudo_count == mpra_data.pseudo_count
    assert data.scaling == mpra_data.scaling


def test_read_and_write_oligo(tmp_path, mpra_oligo_data):
    out_path = tmp_path / "oligo_data.h5ad"
    mpra_oligo_data.write(out_path)

    data = MPRAOligoData.read(out_path)
    assert isinstance(data, MPRAOligoData)
    assert data.data.shape == mpra_oligo_data.data.shape
    assert data.data.layers.keys() == mpra_oligo_data.data.layers.keys()
    assert np.all(data.rna_counts == mpra_oligo_data.rna_counts)
    assert np.all(data.activity == mpra_oligo_data.activity)


def test_read_and_write_with_modifications(tmp_path, mpra_data):
    out_path = tmp_path / "bc_data_mod.h5ad"
    mpra_data.scaling = 10.0
    mpra_data.pseudo_count = 0
    mpra_data.write(out_path)

    data = MPRABarcodeData.read(out_path)
    assert isinstance(data, MPRABarcodeData)
    assert data.scaling == 10.0
    assert data.pseudo_count == 0


def test_modality_from_string():
    assert Modality.from_string("DNA") == Modality.DNA
    assert Modality.from_string("dna") == Modality.DNA
    assert Modality.from_string("RNA") == Modality.RNA
    assert Modality.from_string("rna_normalized") == Modality.RNA_NORMALIZED
    assert Modality.from_string("ACTIVITY") == Modality.ACTIVITY
    with pytest.raises(ValueError):
        Modality.from_string("invalid_modality")


def test_barcodefilter_from_string():
    assert BarcodeFilter.from_string("MIN_BCS_PER_OLIGO") == BarcodeFilter.MIN_BCS_PER_OLIGO
    assert BarcodeFilter.from_string("min_bcs_per_oligo") == BarcodeFilter.MIN_BCS_PER_OLIGO
    assert BarcodeFilter.from_string("GLOBAL") == BarcodeFilter.GLOBAL
    assert BarcodeFilter.from_string("global") == BarcodeFilter.GLOBAL
    assert BarcodeFilter.from_string("OLIGO_SPECIFIC") == BarcodeFilter.OLIGO_SPECIFIC
    assert BarcodeFilter.from_string("LARGE_EXPRESSION") == BarcodeFilter.LARGE_EXPRESSION
    assert BarcodeFilter.from_string("random") == BarcodeFilter.RANDOM
    assert BarcodeFilter.from_string("MIN_COUNT") == BarcodeFilter.MIN_COUNT
    assert BarcodeFilter.from_string("max_count") == BarcodeFilter.MAX_COUNT
    with pytest.raises(ValueError):
        BarcodeFilter.from_string("not_a_filter")


def test_barcode_filter_min_bcs_per_oligo_basic(mpra_data):
    # Should flag all barcodes that have less than the minimum number of barcodes per oligo
    # In the test data, no barcode is a global outlier, so all should be False
    mask = mpra_data._barcode_filter_min_bcs_per_oligo()
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_min_bcs_per_oligo_filter_oligo2(mpra_data):
    # Should flag all barcodes that have less than the minimum number of barcodes per oligo
    # In the test data, oligo2 has only 1 barcode, so it should be flagged when threshold is 2
    mask = mpra_data._barcode_filter_min_bcs_per_oligo(threshold=2)
    expected = np.zeros_like(mask, dtype=bool)
    expected[2, :] = True
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_global_outliers_basic(mpra_data):
    # Should flag barcodes with RNA counts that are global outliers (z-score > 3)
    # In the test data, no barcode is a global outlier, so all should be False
    mask = mpra_data._barcode_filter_global_outliers()
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_global_outliers_basic_with_bc_threshold(mpra_data):
    # Should flag barcodes with RNA counts that are global outliers (z-score > 3)
    # In the test data, no barcode is a global outlier, so all should be False
    mpra_data.barcode_threshold = 2
    mask = mpra_data._barcode_filter_global_outliers(apply_bc_threshold=True)
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_global_outliers_basic_with_bc_threshold_aggregated(mpra_data):
    # Should flag barcodes with RNA counts that are global outliers (z-score > 3)
    # In the test data, no barcode is a global outlier, so all should be False
    mpra_data.data.layers["rna"][0, 0] = 0
    mpra_data.data.layers["dna"][0, 0] = 0
    mpra_data.barcode_threshold = 2
    mask = mpra_data._barcode_filter_global_outliers(times_zscore=0.1, apply_bc_threshold=True)
    expected = np.ones_like(mask, dtype=bool)
    expected[[0, 1], 0] = False  # not applied to oligo1 of rep1 (less than 2 barcodes observed)
    expected[2, :] = False  # not applied to oligo2 (1 barcode)
    mask = mpra_data._barcode_filter_global_outliers(times_zscore=0.1, apply_bc_threshold=True, aggregated_bc_threshold=True)
    expected = np.ones_like(mask, dtype=bool)
    expected[0, 0] = False  # not applied to bcarcode1 of rep1 because not observed
    expected[2, :] = False  # not applied to oligo2 (1 barcode)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_global_outliers_basic_with_bc_threshold_with_outlier(mpra_data):
    # Should flag barcodes with RNA counts that are global outliers (z-score > 3)
    # In the test data, no barcode is a global outlier, so all should be False
    mpra_data.barcode_threshold = 2
    mask = mpra_data._barcode_filter_global_outliers(times_zscore=1.7, apply_bc_threshold=True)
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)
    mask = mpra_data._barcode_filter_global_outliers(times_zscore=1.4, apply_bc_threshold=True)
    expected[4, 2] = True
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_global_outliers_basic_with_outlier(mpra_data):
    # Should flag barcodes with RNA counts that are global outliers (z-score > 3)
    # In the test data, no barcode is a global outlier, so all should be False
    mask = mpra_data._barcode_filter_global_outliers(times_zscore=1.7)
    expected = np.zeros_like(mask, dtype=bool)
    expected[4, 2] = True
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_global_outliers_with_outliers(mpra_data):
    # Introduce a global outlier in RNA counts for a barcode
    mpra_data.data.layers["rna"][1, 0] = 100  # Make barcode2 in rep1 a strong outlier
    mask = mpra_data._barcode_filter_global_outliers(times_zscore=1.7)
    # Only barcode2 in rep1 should be flagged as True
    expected = np.zeros_like(mask, dtype=bool)
    expected[0, 1] = True
    expected[4, 2] = True
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_global_outliers_all_rna_zero(mpra_data):
    # Set all RNA counts to zero, should not raise error and all should be False
    mpra_data.data.layers["rna"][:] = 0
    mask = mpra_data._barcode_filter_global_outliers(times_zscore=1.0)
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_global_outliers_all_zero_true(mpra_data):
    # Set all RNA counts to zero, should not raise error and all should be False
    mpra_data.data.layers["rna"][:] = 0
    mpra_data.data.layers["dna"][:] = 0
    mask = mpra_data._barcode_filter_global_outliers()
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_oligo_specific_outliers_basic(mpra_data):
    # No barcode should be flagged as outlier with high z-score threshold
    mask = mpra_data._barcode_filter_oligo_specific_outliers(times_zscore=3.0)
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_oligo_specific_outliers_with_outlier(mpra_data):
    # change to only 2 oligos becasue we want to have different zscores (at least 3 barcodes needed)
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo3", "oligo3", "oligo3"]
    mask = mpra_data._barcode_filter_oligo_specific_outliers(times_zscore=1.15)
    expected = np.zeros_like(mask, dtype=bool)
    expected[4, 2] = True  # barcode5, rep3
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_oligo_specific_outliers_with_outliers(mpra_data):
    # change to only 2 oligos becasue we want to have different zscores (at least 3 barcodes needed)
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo3", "oligo3", "oligo3"]
    mask = mpra_data._barcode_filter_oligo_specific_outliers(times_zscore=1.0)
    expected = np.zeros_like(mask, dtype=bool)
    expected[4, 2] = True  # barcode5, rep3
    expected[2, 0] = True  # barcode3, rep1
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_oligo_specific_outliers_all_zero(mpra_data):
    # All RNA counts zero, should not raise error and all should be False
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo3", "oligo3", "oligo3"]
    mpra_data.data.layers["rna"][:] = 0
    mask = mpra_data._barcode_filter_oligo_specific_outliers(times_zscore=1.0)
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_oligo_specific_outliers_all_zero_true(mpra_data):
    # All RNA counts zero, should not raise error and all should be False
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo3", "oligo3", "oligo3"]
    mpra_data.data.layers["rna"][:] = 0
    mpra_data.data.layers["dna"][:] = 0
    mask = mpra_data._barcode_filter_oligo_specific_outliers(times_zscore=1.0)
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_large_expression_outliers_basic(mpra_data):
    # No barcode should be flagged as outlier with high z-score threshold
    mask = mpra_data._barcode_filter_large_expression_outliers()
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_large_expression_outliers_with_outlier(mpra_data):
    # change to only 2 oligos becasue we want to have different zscores (at least 3 barcodes needed)
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo3", "oligo3", "oligo3"]
    mask = mpra_data._barcode_filter_large_expression_outliers(times_activity=0.12)
    expected = np.zeros_like(mask, dtype=bool)
    expected[2, :] = True  # barcode3, all replicates
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_large_expression_outliers_with_outliers(mpra_data):
    # change to only 2 oligos becasue we want to have different zscores (at least 3 barcodes needed)
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo3", "oligo3", "oligo3"]
    mpra_data.data.layers["rna"][0, 0] = 100  # Make barcode1 in rep1 a strong outlier
    mask = mpra_data._barcode_filter_large_expression_outliers(times_activity=1.14)
    expected = np.zeros_like(mask, dtype=bool)
    expected[0, :] = True  # barcode1, all replicates
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_large_expression_outliers_all_zero(mpra_data):
    # All RNA counts zero, should not raise error and all should be False
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo3", "oligo3", "oligo3"]
    mpra_data.data.layers["rna"][:] = 0
    mask = mpra_data._barcode_filter_large_expression_outliers(times_activity=0.6)
    expected = np.zeros_like(mask, dtype=bool)
    expected[3, :] = True  # barcode3, all replicates
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_large_expression_outliers_all_zero_true(mpra_data):
    # All RNA counts zero, should not raise error and all should be False
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo3", "oligo3", "oligo3"]
    mpra_data.data.layers["rna"][:] = 0
    mpra_data.data.layers["dna"][:] = 0
    mask = mpra_data._barcode_filter_large_expression_outliers(times_activity=0.6)
    expected = np.zeros_like(mask, dtype=bool)
    np.testing.assert_array_equal(mask, expected)


def test_barcode_filter_combine(mpra_data):
    # change to only 2 oligos because we want to have different zscores (at least 3 barcodes needed)
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo3", "oligo3", "oligo3"]
    mpra_data.apply_barcode_filter(BarcodeFilter.MIN_BCS_PER_OLIGO, {"threshold": 3})
    mpra_data.apply_barcode_filter(BarcodeFilter.OLIGO_SPECIFIC, {"times_zscore": 1.0})
    mask = mpra_data.var_filter
    expected = np.zeros_like(mask, dtype=bool)
    expected[0, :] = True  # barcode1, all replicates
    expected[1, :] = True  # barcode2, all replicates
    expected[4, 2] = True  # barcode5, rep3
    expected[2, 0] = True  # barcode3, rep1
    np.testing.assert_array_equal(mask, expected)


def test_mpra_data_from_file():

    assert MPRAData.from_file("test") is None

    assert MPRAData.from_file("ffasfasdfas") is None


def test_mprabarcode_from_file(tmp_path):
    # Create a mock barcode-level MPRA TSV file
    file_path = tmp_path / "test_bc.tsv"
    df = pd.DataFrame(
        {
            "barcode": ["barcode1", "barcode2", "barcode3"],
            "oligo_name": ["oligo1", "oligo1", "oligo2"],
            "dna_count_rep1": [10, 20, 30],
            "rna_count_rep1": [1, 2, 3],
            "dna_count_rep2": [40, 50, 60],
            "rna_count_rep2": [4, 5, 6],
        }
    )
    df_out = df[["barcode", "oligo_name", "dna_count_rep1", "rna_count_rep1", "dna_count_rep2", "rna_count_rep2"]]
    df_out.to_csv(file_path, sep="\t", index=False)

    # Read using from_file
    data = MPRABarcodeData.from_file(str(file_path))
    assert isinstance(data, MPRABarcodeData)
    assert data.data.shape == (2, 3)  # 2 replicates, 3 barcodes
    assert "rna" in data.data.layers
    assert "dna" in data.data.layers
    assert all(data.data.var["oligo"].isin(["oligo1", "oligo2"]))
    assert data.data.uns["file_path"] == str(file_path)
    assert data.data.uns["normalized"] is False


def test_mpraoligo_from_file(tmp_path):
    # Create a mock AnnData file for oligo-level data
    X = np.array([[1, 2], [3, 4]])
    adata = ad.AnnData(X)
    adata.layers["rna"] = X
    adata.layers["dna"] = X + 10
    adata.var_names = ["oligoA", "oligoB"]
    adata.obs_names = ["rep1", "rep2"]
    file_path = tmp_path / "test_oligo.h5ad"
    adata.write(file_path)

    data = MPRAOligoData.from_file(str(file_path))
    assert isinstance(data, MPRAOligoData)
    assert data.data.shape == (2, 2)
    assert "rna" in data.data.layers
    assert "dna" in data.data.layers
    assert list(data.data.var_names) == ["oligoA", "oligoB"]
    assert list(data.data.obs_names) == ["rep1", "rep2"]


def test_drop_correlation_removes_correlation_layers(mpra_data_norm):
    # Compute correlation to create correlation layers
    mpra_data_norm._normalize()
    mpra_data_norm._compute_correlation(mpra_data_norm.activity, "activity")
    mpra_data_norm._compute_correlation(mpra_data_norm.normalized_rna_counts, "rna_normalized")
    mpra_data_norm._compute_correlation(mpra_data_norm.normalized_dna_counts, "dna_normalized")

    # Ensure correlation layers exist
    for layer in ["activity", "rna_normalized", "dna_normalized"]:
        for method in ["pearson", "spearman"]:
            assert f"{method}_correlation_{layer}" in mpra_data_norm.data.obsp
            assert f"{method}_correlation_{layer}_pvalue" in mpra_data_norm.data.obsp
        assert mpra_data_norm._get_metadata(f"correlation_{layer}") is True

    # Drop correlation
    mpra_data_norm._drop_correlation()

    # Ensure correlation layers are removed and metadata is updated
    for layer in ["activity", "rna_normalized", "dna_normalized"]:
        for method in ["pearson", "spearman"]:
            assert f"{method}_correlation_{layer}" not in mpra_data_norm.data.obsp
            assert f"{method}_correlation_{layer}_pvalue" not in mpra_data_norm.data.obsp
        assert mpra_data_norm._get_metadata(f"correlation_{layer}") is False


def test_drop_correlation_no_error_if_layers_missing(mpra_data_norm):
    # Ensure no correlation layers exist
    for layer in ["activity", "rna_normalized", "dna_normalized"]:
        for method in ["pearson", "spearman"]:
            mpra_data_norm.data.obsp.pop(f"{method}_correlation_{layer}", None)
            mpra_data_norm.data.obsp.pop(f"{method}_correlation_{layer}_pvalue", None)
        mpra_data_norm._add_metadata(f"correlation_{layer}", False)

    # Should not raise error
    mpra_data_norm._drop_correlation()
    # Metadata should remain False
    for layer in ["activity", "rna_normalized", "dna_normalized"]:
        assert mpra_data_norm._get_metadata(f"correlation_{layer}") is False


def test_data_property_getter(mpra_data):
    # Should return the underlying AnnData object
    assert isinstance(mpra_data.data, ad.AnnData)
    # Should have expected shape and layers
    assert mpra_data.data.shape == (3, 5)
    assert "rna" in mpra_data.data.layers
    assert "dna" in mpra_data.data.layers


def test_data_property_setter(mpra_data):
    # Create a new AnnData object and set it
    new_counts = np.ones((3, 5), dtype=np.int32)
    new_obs = pd.DataFrame(index=["repA", "repB", "repC"])
    new_var = pd.DataFrame({"oligo": ["oligoA"] * 5}, index=[f"barcode{i}" for i in range(1, 6)])
    new_layers = {"rna": new_counts.copy(), "dna": new_counts.copy()}
    new_adata = ad.AnnData(X=new_counts.copy(), obs=new_obs.copy(), var=new_var.copy(), layers=new_layers)
    mpra_data.data = new_adata
    assert mpra_data.data is new_adata
    assert mpra_data.data.shape == (3, 5)
    assert np.all(mpra_data.data.layers["rna"] == 1)
    assert np.all(mpra_data.data.layers["dna"] == 1)


@pytest.mark.parametrize(
    "has_sampling",
    [False, True],
)
@pytest.mark.parametrize("modality", ["rna", "dna"])
def test_rna_counts_var_filter_and_sampling(mpra_data, has_sampling, modality):
    # Set up a filter that masks some barcodes
    filter_mask = np.zeros_like(mpra_data.var_filter, dtype=bool)
    filter_mask[1, 0] = True  # mask barcode2, rep1
    filter_mask[3, 2] = True  # mask barcode4, rep3
    mpra_data.var_filter = filter_mask

    if has_sampling:
        # Simulate a sampling layer
        if modality == "rna":
            sampled = mpra_data.raw_rna_counts.copy()
        else:
            sampled = mpra_data.raw_dna_counts.copy()
        sampled[0, 0] = 0
        sampled[2, 1] = 0
        mpra_data.data.layers[f"{modality}_sampling"] = sampled
        expected = sampled * ~filter_mask.T
    else:
        if modality == "rna":
            expected = mpra_data.raw_rna_counts * ~filter_mask.T
        else:
            expected = mpra_data.raw_dna_counts * ~filter_mask.T

    if modality == "rna":
        result = mpra_data.rna_counts
    else:
        result = mpra_data.dna_counts
    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("modality", ["rna", "dna"])
def test_rna_counts_no_var_filter(mpra_data, modality):
    mpra_data.var_filter = None
    if modality == "rna":
        expected = mpra_data.raw_rna_counts
        result = mpra_data.rna_counts
    else:
        expected = mpra_data.raw_dna_counts
        result = mpra_data.dna_counts
    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("modality", ["rna", "dna"])
def test_rna_counts_all_filtered(mpra_data, modality):
    mpra_data.var_filter = np.ones_like(mpra_data.var_filter, dtype=bool)
    if modality == "rna":
        expected = np.zeros_like(mpra_data.raw_rna_counts)
        result = mpra_data.rna_counts
    else:
        expected = np.zeros_like(mpra_data.raw_dna_counts)
        result = mpra_data.dna_counts
    np.testing.assert_array_equal(result, expected)


@pytest.mark.parametrize("modality", ["rna", "dna"])
def test_rna_counts_with_sampling_and_all_filtered(mpra_data, modality):
    mpra_data.var_filter = np.ones_like(mpra_data.var_filter, dtype=bool)
    if modality == "rna":
        sampled = mpra_data.raw_rna_counts.copy()
    else:
        sampled = mpra_data.raw_dna_counts.copy()
    sampled[0, 0] = 99
    mpra_data.data.layers[f"{modality}_sampling"] = sampled
    expected = np.zeros_like(sampled)
    if modality == "rna":
        result = mpra_data.rna_counts
    else:
        result = mpra_data.dna_counts
    np.testing.assert_array_equal(result, expected)


def test_mpra_data_data_property(mpra_data):
    assert isinstance(mpra_data.data, ad.AnnData)
    new_data = copy.deepcopy(mpra_data.data)
    mpra_data.data = new_data
    assert mpra_data.data is new_data


def test_mpra_data_obs_var_names(mpra_data):
    assert list(mpra_data.obs_names) == ["rep1", "rep2", "rep3"]
    assert list(mpra_data.var_names) == ["barcode1", "barcode2", "barcode3", "barcode4", "barcode5"]
    assert mpra_data.n_obs == 3
    assert mpra_data.n_vars == 5


def test_mpra_data_oligos(mpra_data):
    oligos = mpra_data.oligos
    assert isinstance(oligos, pd.Series)
    assert set(oligos) == {"oligo1", "oligo2", "oligo3"}


def test_mpra_data_raw_counts(mpra_data):
    assert np.array_equal(mpra_data.raw_rna_counts, COUNTS_RNA)
    assert np.array_equal(mpra_data.raw_dna_counts, COUNTS_DNA)


def test_mpra_data_total_counts(mpra_data):
    # Should match sum over axis 1
    assert np.array_equal(mpra_data.total_rna_counts, COUNTS_RNA.sum(axis=1))
    assert np.array_equal(mpra_data.total_dna_counts, COUNTS_DNA.sum(axis=1))


def test_mpra_data_drop_total_counts(mpra_data):
    mpra_data.total_rna_counts  # populate
    mpra_data.total_dna_counts
    mpra_data.drop_total_counts()
    assert "rna_counts" not in mpra_data.data.obs
    assert "dna_counts" not in mpra_data.data.obs


def test_mpra_data_var_filter_setter_and_getter(mpra_data):
    mpra_data.var_filter = FILTER
    assert np.array_equal(mpra_data.var_filter, FILTER)
    mpra_data.var_filter = None
    assert np.all(mpra_data.var_filter == False)  # noqa: E712


def test_mpra_data_add_sequence_design(mpra_data):
    # Create a minimal sequence design DataFrame
    df = pd.DataFrame(
        {
            "category": ["cat1", "cat2", "cat3"],
            "class": ["class1", "class2", "class3"],
            "ref": ["ref1", "ref2", "ref3"],
            "chr": ["chr1", "chr2", "chr3"],
            "start": [1, 2, 3],
            "end": [10, 20, 30],
            "strand": ["+", "-", "+"],
            "variant_class": ["vc1", "vc2", "vc3"],
            "variant_pos": [100, 200, 300],
            "SPDI": [["spdi1"], ["spdi2"], ["spdi3"]],
            "allele": [["ref"], ["alt"], ["ref"]],
        },
        index=["oligo1", "oligo2", "oligo3"],
    )
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo2", "oligo3", "oligo3"]
    mpra_data.add_sequence_design(df, "dummy_path")
    assert "category" in mpra_data.data.var
    assert mpra_data._get_metadata("sequence_design_file") == "dummy_path"


def test_mpra_data_variant_map(mpra_data):
    # Setup sequence design
    df = pd.DataFrame(
        {
            "category": ["cat1", "cat2", "cat3"],
            "class": ["class1", "class2", "class3"],
            "ref": ["ref1", "ref2", "ref3"],
            "chr": ["chr1", "chr2", "chr3"],
            "start": [1, 2, 3],
            "end": [10, 20, 30],
            "strand": ["+", "-", "+"],
            "variant_class": ["vc1", "vc2", "vc3"],
            "variant_pos": [100, 200, 300],
            "SPDI": [["spdi1"], ["spdi2"], ["spdi3"]],
            "allele": [["ref"], ["alt"], ["ref"]],
        },
        index=["oligo1", "oligo2", "oligo3"],
    )
    mpra_data.data.var["oligo"] = ["oligo1", "oligo1", "oligo2", "oligo3", "oligo3"]
    mpra_data.add_sequence_design(df, "dummy_path")
    variant_map = mpra_data.variant_map
    assert isinstance(variant_map, pd.DataFrame)
    assert "REF" in variant_map.columns
    assert "ALT" in variant_map.columns


@pytest.mark.parametrize("modality", ["rna", "dna"])
def test_mpra_data_drop_normalized(mpra_data, modality):
    mpra_data._normalize()
    assert f"{modality}_normalized" in mpra_data.data.layers
    mpra_data.drop_normalized()
    assert f"{modality}_normalized" not in mpra_data.data.layers
    assert mpra_data._get_metadata("normalized") is False


@pytest.mark.parametrize("modality", ["rna", "dna"])
def test_mpra_data_drop_correlation(mpra_data, modality):
    mpra_data._normalize()
    if modality == "rna":
        mpra_data._compute_correlation(mpra_data.normalized_rna_counts, "rna_normalized")
    else:
        mpra_data._compute_correlation(mpra_data.normalized_dna_counts, "dna_normalized")
    assert mpra_data._get_metadata(f"correlation_{modality}_normalized") is True
    mpra_data._drop_correlation()
    assert mpra_data._get_metadata(f"correlation_{modality}_normalized") is False


def test_mpra_data_correlation_invalid_method(mpra_data):
    mpra_data._normalize()
    with pytest.raises(ValueError):
        mpra_data.correlation(method="invalid", count_type=Modality.RNA_NORMALIZED)
    with pytest.raises(ValueError):
        mpra_data.correlation(method="pearson", count_type=Modality.RNA)


def test_drop_count_sampling_removes_sampling_layers_and_metadata(mpra_data):
    # Apply count sampling to create sampling layers and metadata
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, proportion=0.5)
    assert "rna_sampling" in mpra_data.data.layers
    assert "dna_sampling" in mpra_data.data.layers
    assert "count_sampling" in mpra_data.data.uns

    # Drop count sampling
    mpra_data.drop_count_sampling()

    # Sampling layers and metadata should be removed
    assert "rna_sampling" not in mpra_data.data.layers
    assert "dna_sampling" not in mpra_data.data.layers
    assert "count_sampling" not in mpra_data.data.uns


def test_drop_count_sampling_also_drops_normalized_and_barcode_counts(mpra_data):
    # Apply count sampling and normalization
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, proportion=0.5)
    mpra_data._normalize()
    mpra_data.barcode_counts = np.ones_like(mpra_data.raw_rna_counts)

    # Confirm layers exist
    assert "rna_normalized" in mpra_data.data.layers
    assert "dna_normalized" in mpra_data.data.layers
    assert "barcode_counts" in mpra_data.data.layers

    # Drop count sampling
    mpra_data.drop_count_sampling()

    # Normalized and barcode counts layers should be removed
    assert "rna_normalized" not in mpra_data.data.layers
    assert "dna_normalized" not in mpra_data.data.layers
    assert "barcode_counts" not in mpra_data.data.layers


def test_drop_count_sampling_removes_total_counts(mpra_data):
    # Apply count sampling to create total counts
    np.random.seed(42)
    mpra_data.apply_count_sampling(CountSampling.RNA_AND_DNA, proportion=0.5)
    _ = mpra_data.total_rna_counts
    _ = mpra_data.total_dna_counts
    assert "rna_counts" in mpra_data.data.obs
    assert "dna_counts" in mpra_data.data.obs

    # Drop count sampling
    mpra_data.drop_count_sampling()

    # Total counts should be removed
    assert "rna_counts" not in mpra_data.data.obs
    assert "dna_counts" not in mpra_data.data.obs


def test_modality_enum():
    assert Modality.DNA.value == "dna"
    assert Modality.RNA.value == "rna"
    assert Modality.DNA_NORMALIZED.value == "dna_normalized"
    assert Modality.RNA_NORMALIZED.value == "rna_normalized"
    assert Modality.ACTIVITY.value == "activity"
    assert Modality.from_string("DNA") == Modality.DNA
    assert Modality.from_string("rna_normalized") == Modality.RNA_NORMALIZED
    with pytest.raises(ValueError):
        Modality.from_string("not_a_modality")


def test_countsampling_enum():
    assert CountSampling.RNA.value == "RNA"
    assert CountSampling.DNA.value == "DNA"
    assert CountSampling.RNA_AND_DNA.value == "RNA_AND_DNA"


def test_barcodefilter_enum():
    assert BarcodeFilter.MIN_BCS_PER_OLIGO.value == "min_bcs_per_oligo"
    assert BarcodeFilter.GLOBAL.value == "global"
    assert BarcodeFilter.OLIGO_SPECIFIC.value == "oligo_specific"
    assert BarcodeFilter.LARGE_EXPRESSION.value == "large_expression"
    assert BarcodeFilter.RANDOM.value == "random"
    assert BarcodeFilter.MIN_COUNT.value == "min_count"
    assert BarcodeFilter.MAX_COUNT.value == "max_count"
    assert BarcodeFilter.from_string("MIN_COUNT") == BarcodeFilter.MIN_COUNT
    with pytest.raises(ValueError):
        BarcodeFilter.from_string("not_a_filter")


def test_mpra_data_scaling_and_pseudocount(mpra_data):
    old_scaling = mpra_data.scaling
    mpra_data.scaling = 12345.0
    assert mpra_data.scaling == 12345.0
    mpra_data.scaling = old_scaling
    old_pseudocount = mpra_data.pseudo_count
    mpra_data.pseudo_count = 7
    assert mpra_data.pseudo_count == 7
    mpra_data.pseudo_count = old_pseudocount


def test_mpra_data_metadata(mpra_data):
    mpra_data._add_metadata("test_key", "test_value")
    assert mpra_data._get_metadata("test_key") == "test_value"


def test_mpra_data_var_filter_setter(mpra_data):
    mpra_data.var_filter = None
    assert np.all(mpra_data.var_filter == False)  # noqa: E712
    mask = np.ones_like(mpra_data.var_filter, dtype=bool)
    mpra_data.var_filter = mask
    assert np.all(mpra_data.var_filter == True)  # noqa: E712


def test_mpra_data_write_and_read(tmp_path, mpra_data):
    out_path = tmp_path / "test_write.h5ad"
    mpra_data.write(out_path)
    loaded = MPRABarcodeData.read(out_path)
    assert isinstance(loaded, MPRABarcodeData)
    assert loaded.data.shape == mpra_data.data.shape


def test_mpraoligodata_from_file(tmp_path, mpra_oligo_data):
    out_path = tmp_path / "oligo_data_test.h5ad"
    mpra_oligo_data.write(out_path)
    loaded = MPRAOligoData.from_file(out_path)
    assert isinstance(loaded, MPRAOligoData)
    assert loaded.data.shape == mpra_oligo_data.data.shape


def test_mpraoligodata_barcode_counts_exception(mpra_oligo_data):
    mpra_oligo_data.data.layers.pop("barcode_counts", None)
    with pytest.raises(MPRAlibException):
        _ = mpra_oligo_data.barcode_counts


def test_mpradata_normalize_layer(mpra_data):
    norm = mpra_data._normalize_layer(mpra_data.dna_counts, mpra_data.total_dna_counts)
    assert norm.shape == mpra_data.dna_counts.shape
    assert np.all(norm >= 0)


def test_mpradata_drop_normalized(mpra_data):
    mpra_data._normalize()
    assert "rna_normalized" in mpra_data.data.layers
    mpra_data.drop_normalized()
    assert "rna_normalized" not in mpra_data.data.layers
    assert "dna_normalized" not in mpra_data.data.layers


def test_mpradata_drop_total_counts(mpra_data):
    mpra_data.total_rna_counts
    mpra_data.total_dna_counts
    mpra_data.drop_total_counts()
    assert "rna_counts" not in mpra_data.data.obs
    assert "dna_counts" not in mpra_data.data.obs


def test_mprabarcode_complexity(mpra_data):
    result = mpra_data.complexity()
    assert result.shape[0] == mpra_data.n_obs
    assert result.shape[1] == mpra_data.n_obs


def test_mprabarcode_apply_count_sampling(mpra_data):
    mpra_data.apply_count_sampling(CountSampling.RNA, proportion=0.5)
    assert "rna_sampling" in mpra_data.data.layers
    mpra_data.apply_count_sampling(CountSampling.DNA, proportion=0.5)
    assert "dna_sampling" in mpra_data.data.layers


def test_mprabarcode_apply_barcode_filter(mpra_data):
    mpra_data.apply_barcode_filter(BarcodeFilter.RANDOM, params={"proportion": 0.5})
    assert mpra_data.var_filter.shape == (mpra_data.n_vars, mpra_data.n_obs)
    mpra_data.apply_barcode_filter(BarcodeFilter.MIN_COUNT, params={"rna_min_count": 1})
    assert mpra_data.var_filter.shape == (mpra_data.n_vars, mpra_data.n_obs)


def test_mprabarcode_drop_count_sampling(mpra_data):
    mpra_data.apply_count_sampling(CountSampling.RNA, proportion=0.5)
    mpra_data.drop_count_sampling()
    assert "rna_sampling" not in mpra_data.data.layers
    assert "dna_sampling" not in mpra_data.data.layers


def test_mprabarcode_drop_barcode_counts(mpra_data):
    mpra_data.barcode_counts = np.ones_like(mpra_data.raw_rna_counts)
    mpra_data.drop_barcode_counts()
    assert "barcode_counts" not in mpra_data.data.layers


def test_mpraoligodata_normalize_layer(mpra_oligo_data):
    counts = np.array([[1, 2, 3], [3, 4, 5], [3, 4, 5]], dtype=np.int32)
    total_counts = np.array([3, 7, 7], dtype=np.int32)
    mpra_oligo_data.barcode_counts = np.ones_like(counts)
    norm = mpra_oligo_data._normalize_layer(counts, total_counts)
    assert norm.shape == counts.shape
    assert np.all(norm >= 0)



import copy
import gzip
import os
import tempfile
import pandas as pd
import numpy as np
from numpy.typing import NDArray
from typing import Optional
import pytest
import anndata as ad
from mpralib.utils.io import export_barcode_file, read_sequence_design_file, export_counts_file, chromosome_map
from mpralib.exception import SequenceDesignException, MPRAlibException
from mpralib.mpradata import MPRABarcodeData, MPRAOligoData, MPRAData


class DummyMPRAData(MPRAData):
    def __init__(
        self,
        replicates: list[str],
        oligos: list[str],
        barcodes: list[str],
        dna_counts: NDArray[np.int32],
        rna_counts: NDArray[np.int32],
        barcode_threshold: int,
        barcode_counts: Optional[NDArray[np.int32]] = None,
    ):
        layers = {"rna": rna_counts, "dna": dna_counts}
        obs = pd.DataFrame(index=replicates)
        var = pd.DataFrame({"oligo": oligos}, index=barcodes)
        super().__init__(ad.AnnData(X=rna_counts, obs=obs, var=var, layers=layers), barcode_threshold)
        if barcode_counts is not None:
            self.barcode_counts = barcode_counts


class DummyMPRABarcodeData(DummyMPRAData, MPRABarcodeData):
    pass


class DummyMPRAOligoData(DummyMPRAData, MPRAOligoData):
    pass


@pytest.fixture
def barcode_data():
    replicates = ["rep1", "rep2"]
    barcodes = ["bc1", "bc2", "bc3"]
    oligos = ["ol1", "ol2", "ol3"]
    dna_counts = np.array([[10, 0, 5], [20, 1, 0]])
    rna_counts = np.array([[5, 0, 2], [10, 1, 0]])
    barcode_threshold = 1
    return DummyMPRABarcodeData(replicates, oligos, barcodes, dna_counts, rna_counts, barcode_threshold)


@pytest.fixture
def oligo_data():
    replicates = ["rep1", "rep2"]
    barcodes = ["bc1", "bc2", "bc3"]
    oligos = ["ol1", "ol2", "ol3"]
    dna_counts = np.array([[10, 0, 5], [20, 1, 0]], dtype=np.int32)
    rna_counts = np.array([[5, 0, 2], [10, 1, 0]], dtype=np.int32)
    barcode_counts = np.array([[1, 0, 1], [1, 0, 1]], dtype=np.int32)
    barcode_threshold = 1
    return DummyMPRAOligoData(replicates, oligos, barcodes, dna_counts, rna_counts, barcode_threshold, barcode_counts)


def test_export_counts_file_barcode(tmp_path, barcode_data):
    out_path = tmp_path / "counts.tsv"
    export_counts_file(barcode_data, str(out_path), normalized=False)
    df = pd.read_csv(out_path, sep="\t", na_values=["X"])
    # Only rows with observed=True and barcode_counts >= threshold should remain
    assert "oligo_name" in df.columns
    assert set(df["oligo_name"]) == {"ol1", "ol2", "ol3"}
    assert "dna_count_rep1" in df.columns
    assert "rna_count_rep2" in df.columns
    assert "rna_count_rep2" in df.columns
    assert df.loc[df["oligo_name"] == "ol3", "dna_count_rep1"].values[0] == 5
    # Check that zeros are written as empty
    assert np.isnan(df.loc[df["oligo_name"] == "ol3", "dna_count_rep2"].values[0])


def test_export_counts_file_oligo(tmp_path, oligo_data):
    out_path = tmp_path / "counts.tsv"
    export_counts_file(oligo_data, str(out_path), normalized=False)
    df = pd.read_csv(out_path, sep="\t")
    # Only rows with observed=True and barcode_counts >= threshold should remain
    assert "ID" in df.columns
    assert set(df["ID"]) == {"ol1", "ol3"}
    assert "dna_count_rep1" in df.columns
    assert "rna_count_rep2" in df.columns


def test_export_counts_file_normalized(tmp_path, barcode_data):
    out_path = tmp_path / "counts_norm.tsv"
    export_counts_file(barcode_data, str(out_path), normalized=True)
    df = pd.read_csv(out_path, sep="\t")
    # Check float formatting
    assert all(df.filter(like="dna_count").apply(lambda x: isinstance(x, float) or pd.isna(x)))


def test_export_counts_file_with_filter(tmp_path, barcode_data):
    out_path = tmp_path / "counts_filt.tsv"
    # Mask out the first entry
    filter_mask = np.asarray([[True, False, False], [True, False, False]])
    export_counts_file(barcode_data, str(out_path), normalized=False, filter=filter_mask)
    df = pd.read_csv(out_path, sep="\t")
    # Only ol3 should remain (ol2 fails barcode_threshold, ol1 is masked)
    assert set(df["oligo_name"]) == {"ol2", "ol3"}


def test_export_counts_file_invalid_type(tmp_path):
    class Dummy:
        pass

    dummy = Dummy()
    with pytest.raises(MPRAlibException):
        export_counts_file(dummy, str(tmp_path / "fail.tsv"))  # type: ignore


VALID_TSV = """
name\tsequence\tcategory\tclass\tsource\tref\tchr\tstart\tend\tstrand\tvariant_class\tvariant_pos\tSPDI\tallele\tinfo
oligo1\tATGCATGC\tvariant\ttest\tsrc1\tref1\tchr1\t100\t200\t+\t["SNV"]\t[123]\t["NC_00007.14:117548628:TTTTTTT:TTTTTTTTT"]\t["ref"]\tinfo1
oligo2\tGGCATGCA\telement\tvariant positive control\tsrc2\tref2\tchr2\t300\t400\t-\tNA\tNA\tNA\tNA\tinfo2
oligo3\tATGCATGC\tvariant\ttest\tsrc1\tref1\tchr1\t100\t200\t+\t["SNV"]\t[123]\t["NC_00007.14:117548628:TTTTTTT:TTTTTTTTT"]\t["alt"]\tinfo1
"""

INVALID_SEQ_TSV = """
name\tsequence\tcategory\tclass\tsource\tref\tchr\tstart\tend\tstrand\tvariant_class\tvariant_pos\tSPDI\tallele\tinfo
oligo1\tATGCXTGC\tvariant\ttest\tsrc1\tref1\tchr1\t100\t200\t+\t["SNV"]\t[123]\t["NC_00007.14:117548628:TTTTTTT:TTTTTTTTT"]\t["alt"]\tinfo1
"""

INVALID_CATEGORY_TSV = """
name\tsequence\tcategory\tclass\tsource\tref\tchr\tstart\tend\tstrand\tvariant_class\tvariant_pos\tSPDI\tallele\tinfo
oligo1\tATGCATGC\tbadcat\ttest\tsrc1\tref1\tchr1\t100\t200\t+\tNA\tNA\tNA\tNA\tinfo1
"""

INVALID_CLASS_TSV = """
name\tsequence\tcategory\tclass\tsource\tref\tchr\tstart\tend\tstrand\tvariant_class\tvariant_pos\tSPDI\tallele\tinfo
oligo1\tATGCATGC\tvariant\tbadclass\tsrc1\tref1\tchr1\t100\t200\t+\t["SNV"]\t[123]\t["NC_00007.14:117548628:TTTTTTT:TTTTTTTTT"]\t["alt"]\tinfo1
"""


def write_temp_tsv(content):
    fd, path = tempfile.mkstemp(suffix=".tsv")
    with os.fdopen(fd, "w") as f:
        f.write(content)
    return path


def test_read_sequence_design_file_valid():
    path = write_temp_tsv(VALID_TSV)
    df = read_sequence_design_file(path)
    assert isinstance(df, pd.DataFrame)
    assert "sequence" in df.columns
    assert df.shape[0] == 3
    assert set(df.index) == {"oligo1", "oligo2", "oligo3"}
    os.remove(path)


def test_read_sequence_design_file_invalid_sequence():
    path = write_temp_tsv(INVALID_SEQ_TSV)
    with pytest.raises(SequenceDesignException) as excinfo:
        read_sequence_design_file(path)
    assert "sequence" in str(excinfo.value)
    os.remove(path)


def test_read_sequence_design_file_invalid_category():
    path = write_temp_tsv(INVALID_CATEGORY_TSV)
    with pytest.raises(SequenceDesignException) as excinfo:
        read_sequence_design_file(path)
    assert "category" in str(excinfo.value)
    os.remove(path)


def test_read_sequence_design_file_invalid_class():
    path = write_temp_tsv(INVALID_CLASS_TSV)
    with pytest.raises(SequenceDesignException) as excinfo:
        read_sequence_design_file(path)
    assert "class" in str(excinfo.value)
    os.remove(path)


def test_chromosome_map_columns():
    df = chromosome_map()
    expected_columns = ["ucsc", "assembly", "genbank", "refseq", "release"]
    assert list(df.columns) == expected_columns


def test_chromosome_map_release_values():
    df = chromosome_map()
    assert set(df["release"].unique()) == {"GRCh37", "GRCh38"}


def test_chromosome_map_non_empty():
    df = chromosome_map()
    assert not df.empty
    assert df.shape[0] > 0


def test_chromosome_map_no_duplicate_rows():
    df = chromosome_map()
    assert df.duplicated().sum() == 0


OBS = pd.DataFrame(index=["rep1", "rep2", "rep3"])
VAR = pd.DataFrame(
    {"oligo": ["oligo1", "oligo1", "oligo2", "oligo3", "oligo3"]},
    index=["barcode1", "barcode2", "barcode3", "barcode4", "barcode5"],
)
COUNTS_DNA = np.array([[1, 2, 3, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])
COUNTS_RNA = np.array([[1, 2, 4, 1, 2], [4, 5, 6, 4, 5], [7, 8, 9, 10, 100]])

FILTER = np.array(
    [
        [False, True, False],
        [False, False, False],
        [True, False, False],
        [False, False, True],
        [False, False, True],
    ]
)


@pytest.fixture
def mpra_data():
    layers = {"rna": COUNTS_RNA.copy(), "dna": COUNTS_DNA.copy()}
    return MPRABarcodeData(ad.AnnData(X=COUNTS_RNA.copy(), obs=OBS.copy(), var=VAR.copy(), layers=layers))


@pytest.fixture
def mpra_data_with_bc_filter(mpra_data):
    data = copy.deepcopy(mpra_data)
    data.var_filter = FILTER
    return data


@pytest.fixture
def files():
    input_file = os.path.join(os.path.dirname(__file__), "data", "reporter_experiment_barcode.input.tsv.gz")
    output_file_activity = tempfile.NamedTemporaryFile(delete=False).name
    output_file_barcode = tempfile.NamedTemporaryFile(delete=False).name
    yield {"input": input_file, "output_activity": output_file_activity, "output_barcode": output_file_barcode}
    os.remove(output_file_activity)
    os.remove(output_file_barcode)


def test_export_barcode_file(mpra_data, files):
    export_barcode_file(mpra_data, files["output_barcode"])
    assert os.path.exists(files["output_barcode"])

    with open(files["output_barcode"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(
        os.path.dirname(__file__), "data", "io", "reporter_experiment_barcode.simple.out.tsv.gz"
    )

    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content


def test_export_barcode_file_var_filter(mpra_data_with_bc_filter, files):
    export_barcode_file(mpra_data_with_bc_filter, files["output_barcode"])
    assert os.path.exists(files["output_barcode"])

    with open(files["output_barcode"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(
        os.path.dirname(__file__), "data", "io", "reporter_experiment_barcode.simple_with_bc_filter.out.tsv.gz"
    )

    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content



import pytest
import os
import tempfile
from click.testing import CliRunner
from mpralib.cli import cli, _get_chr
import gzip
import pandas as pd
from logging import Logger


@pytest.fixture(scope="module")
def runner():
    return CliRunner()


def test_cli_group_help(runner):
    result = runner.invoke(cli, ["--help"])
    assert result.exit_code == 0
    assert "Command line interface of MPRAlib" in result.output


@pytest.fixture
def files():
    input_file = os.path.join(os.path.dirname(__file__), "data", "reporter_experiment_barcode.input.tsv.gz")
    output_file = tempfile.NamedTemporaryFile(delete=False).name
    yield {"input": input_file, "output": output_file}
    os.remove(output_file)


def test_barcode_activities_bc1(runner, files):

    # Run the command
    result = runner.invoke(
        cli,
        [
            "functional",
            "activities",
            "--input",
            files["input"],
            "--barcode-level",
            "--output",
            files["output"],
        ],
    )

    # Check the result
    assert result.exit_code == 0
    assert os.path.exists(files["output"])

    with open(files["output"], "r") as f:
        output_content = f.read()

    expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_experiment_barcode.input.tsv.gz")

    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()

    assert output_content == expected_content


def test_activities_bc1(runner, files):

    # Run the command
    result = runner.invoke(
        cli,
        [
            "functional",
            "activities",
            "--input",
            files["input"],
            "--bc-threshold",
            "1",
            "--output",
            files["output"],
        ],
    )

    # Check the result
    assert result.exit_code == 0
    assert os.path.exists(files["output"])

    with open(files["output"], "r") as f:
        output_content = f.read()

    expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_activity.bc1.output.tsv.gz")

    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()

    assert output_content == expected_content


def test_activities_bc10(runner, files):

    # Run the command
    result = runner.invoke(
        cli,
        [
            "functional",
            "activities",
            "--input",
            files["input"],
            "--bc-threshold",
            "10",
            "--output",
            files["output"],
        ],
    )

    # Check the result
    assert result.exit_code == 0
    assert os.path.exists(files["output"])

    with open(files["output"], "r") as f:
        output_content = f.read()

    expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_activity.bc10.output.tsv.gz")

    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()

    assert output_content == expected_content


def test_activities_bc100(runner, files):

    # Run the command
    result = runner.invoke(
        cli,
        [
            "functional",
            "activities",
            "--input",
            files["input"],
            "--bc-threshold",
            "100",
            "--output",
            files["output"],
        ],
    )

    # Check the result
    assert result.exit_code == 0
    assert os.path.exists(files["output"])

    with open(files["output"], "r") as f:
        output_content = f.read()

    expected_output_file = os.path.join(os.path.dirname(__file__), "data", "reporter_activity.bc100.output.tsv.gz")

    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()

    assert output_content == expected_content


class DummyLogger(Logger):

    def __init__(self):
        self.messages = []

    def warning(self, msg, *args, **kwargs):
        self.messages.append(msg)


@pytest.fixture
def logger():
    return DummyLogger()


def test_get_chr_found(logger):
    # Prepare a chromosome map DataFrame
    map_df = pd.DataFrame(
        {"refseq": ["NC_000001.11", "NC_000002.12"], "ucsc": ["chr1", "chr2"], "release": ["GRCh38", "GRCh38"]}
    )
    variant_id = "NC_000001.11:12345:A:T"
    result = _get_chr(map_df, variant_id, logger)
    assert result == "chr1"
    assert logger.messages == []


def test_get_chr_not_found(logger):
    map_df = pd.DataFrame(
        {"refseq": ["NC_000001.11", "NC_000002.12"], "ucsc": ["chr1", "chr2"], "release": ["GRCh38", "GRCh38"]}
    )
    variant_id = "NC_000003.13:54321:G:C"
    result = _get_chr(map_df, variant_id, logger)
    assert result is None
    assert any("Contig NC_000003.13 of SPDI NC_000003.13:54321:G:C not found" in msg for msg in logger.messages)


def test_get_chr_handles_empty_map(logger):
    map_df = pd.DataFrame(columns=["refseq", "ucsc", "release"])
    variant_id = "NC_000004.14:11111:T:A"
    result = _get_chr(map_df, variant_id, logger)
    assert result is None
    assert any("Contig NC_000004.14 of SPDI NC_000004.14:11111:T:A not found" in msg for msg in logger.messages)


def test_get_chr_with_multiple_matches(logger):
    # Should return the first match if multiple rows match
    map_df = pd.DataFrame(
        {"refseq": ["NC_000005.15", "NC_000005.15"], "ucsc": ["chr5a", "chr5b"], "release": ["GRCh38", "GRCh37"]}
    )
    variant_id = "NC_000005.15:22222:C:G"
    result = _get_chr(map_df, variant_id, logger)
    assert result in ["chr5a", "chr5b"]
    assert logger.messages == []



import os
import tempfile
import pytest
from click.testing import CliRunner
import gzip
from mpralib.cli import cli


@pytest.fixture(scope="module")
def runner():

    return CliRunner()


def test_combine_group_exists(runner):
    result = runner.invoke(cli, ["combine", "--help"])
    assert result.exit_code == 0
    assert "Combine counts with other outputs." in result.output


def test_combine_get_counts_option_required(runner):
    result = runner.invoke(cli, ["combine", "get-counts"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


def test_combine_get_reporter_elements_option_required(runner):
    result = runner.invoke(cli, ["combine", "get-reporter-elements"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


def test_combine_get_reporter_variants_option_required(runner):
    result = runner.invoke(cli, ["combine", "get-reporter-variants"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


def test_combine_get_reporter_genomic_elements_option_required(runner):
    result = runner.invoke(cli, ["combine", "get-reporter-genomic-elements"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


def test_combine_get_reporter_genomic_variants_option_required(runner):
    result = runner.invoke(cli, ["combine", "get-reporter-genomic-variants"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


def test_combine_get_variant_counts_option_required(runner):
    result = runner.invoke(cli, ["combine", "get-variant-counts"])
    assert result.exit_code != 0
    assert "Missing option '--input'" in result.output


def test_combine_get_variant_map_option_required(runner):
    result = runner.invoke(cli, ["combine", "get-variant-map"])
    assert result.exit_code != 0
    assert "Missing option '--sequence-design'" in result.output


@pytest.fixture
def files():
    base = os.path.dirname(__file__)
    output_file = tempfile.NamedTemporaryFile(delete=False).name
    yield {
        "input": {
            "barcode_counts": os.path.join(base, "data", "reporter_experiment_barcode_IGVFDS2165KBMD.input.head20000.tsv.gz"),
            "sequence_design": os.path.join(base, "data", "reporter_sequence_design.example.tsv.gz"),
        },
        "output": output_file,
    }
    os.remove(output_file)


def test_combine_get_counts_oligos(runner, files):
    result = runner.invoke(
        cli,
        [
            "combine",
            "get-counts",
            "--input",
            files["input"]["barcode_counts"],
            "--sequence-design",
            files["input"]["sequence_design"],
            "--output",
            files["output"],
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists(files["output"])

    with open(files["output"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(
        os.path.dirname(__file__), "data", "combine", "oligo_counts.IGVFDS2165KBMD.head20000.output.tsv.gz"
    )

    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content


def test_combine_get_counts_barcodes(runner, files):
    result = runner.invoke(
        cli,
        [
            "combine",
            "get-counts",
            "--barcodes",
            "--input",
            files["input"]["barcode_counts"],
            "--sequence-design",
            files["input"]["sequence_design"],
            "--output",
            files["output"],
        ],
    )
    assert result.exit_code == 0
    assert os.path.exists(files["output"])

    with open(files["output"], "r") as f:
        output_content = f.read()
    expected_output_file = files["input"]["barcode_counts"]

    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    expected_content = "ID" + expected_content[7:]
    assert output_content == expected_content


def test_combine_get_variant_map(runner, files):
    result = runner.invoke(
        cli,
        [
            "combine",
            "get-variant-map",
            "--sequence-design",
            files["input"]["sequence_design"],
            "--output",
            files["output"],
        ],
    )

    assert result.exit_code == 0
    assert os.path.exists(files["output"])

    with open(files["output"], "r") as f:
        output_content = f.read()
    expected_output_file = os.path.join(os.path.dirname(__file__), "data", "combine", "get_variant_map.output.tsv.gz")

    with gzip.open(expected_output_file, "rt") as f:
        expected_content = f.read()
    assert output_content == expected_content



from setuptools import setup, find_packages

setup(
    name="mpralib",
    version="0.9.0",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    include_package_data=True,
    package_data={
        "MPRAlib": ["data/*.txt"],
    },
    install_requires=[
        "numpy",
        "pandas",
        "pysam",
        "click",
        "scikit-learn",
        "scipy",
        "anndata>=0.11.3",
        "seaborn",
        "matplotlib",
        "jsonschema",
        "tqdm",
    ],
    extras_require={
        "test": [
            "pytest",
            "coverage",
        ],
    },
    author="Max Schubach",
    author_email="max.schubach@bih-charite.de",
    description="Library to analyze count data of MPRA experiments.",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    url="https://github.com/kircherlab/MPRAlib",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.9",
    entry_points={"console_scripts": ["mpralib=mpralib.cli:main"]},
)
